import sys
from abc import ABC, abstractmethod
from typing import Optional


class PythonVersionChecker(ABC):
    """
    Interface for classes which check if the Python version matches a certain criteria.
    """
    @abstractmethod
    def check(self, major: int, minor: int, micro: int, releaselevel: str, serial: int) -> Optional[str]:
        """
        Checks if the given Python version is supported. Returns None if it is supported,
        or an error message if not.

        :param major:           The major version number.
        :param minor:           The minor version number.
        :param micro:           The micro version number.
        :param releaselevel:    The release-level string.
        :param serial:          The serial number.
        :return:                None for pass and error-string for fail.
        """
        pass

    def check_current(self) -> Optional[str]:
        """
        Checks the current Python version.

        :return:    None for pass and error-string for fail.
        """
        # Get the current Python version
        version = sys.version_info

        # Run the check
        return self.check(version.major,
                          version.minor,
                          version.micro,
                          version.releaselevel,
                          version.serial)

    def ensure_current(self):
        """
        Ensures the current Python version passes this check, raising an error if not.
        """
        # Check the current Python version
        error: Optional[str] = self.check_current()

        # If the check failed, raise
        if error is not None:
            raise PythonVersionError(f"Python version check failed: {error}")

    def current_passes(self) -> bool:
        """
        Checks if the current Python version passes this check.
        """
        return self.check_current() is None


class PythonVersionError(Exception):
    """
    Error class for incorrect Python versions.
    """
    pass



from typing import Tuple, Optional

from ._Point import Point


class Rectangle:
    """
    Represents an axis-aligned rectangle in 2D integer Cartesian coordinates.
    """
    def __init__(self, p1: Point, p2: Point):
        self.p1: Point = p1
        self.p2: Point = p2

    def left(self) -> int:
        """
        Gets the x-coordinate of the left-hand side of this rectangle.

        :return:    The x-coordinate.
        """
        return min(self.p1.x, self.p2.x)

    def right(self) -> int:
        """
        Gets the x-coordinate of the right-hand side of this rectangle.

        :return:    The x-coordinate.
        """
        return max(self.p1.x, self.p2.x)

    def top(self) -> int:
        """
        Gets the y-coordinate of the top side of this rectangle.

        :return:    The y-coordinate.
        """
        return min(self.p1.y, self.p2.y)

    def bottom(self) -> int:
        """
        Gets the y-coordinate of the bottom side of this rectangle.

        :return:    The y-coordinate.
        """
        return max(self.p1.y, self.p2.y)

    def width(self) -> int:
        """
        Gets the width of the rectangle.

        :return: The rectangle's width.
        """
        return abs(self.p1.x - self.p2.x) + 1

    def height(self) -> int:
        """
        Gets the height of the rectangle.

        :return:    The rectangle's height.
        """
        return abs(self.p1.y - self.p2.y) + 1

    def area(self) -> int:
        """
        Gets the area of this rectangle.

        :return:    The area.
        """
        return self.width() * self.height()

    @classmethod
    def x_y_w_h(cls, x: int, y: int, w: int, h: int) -> "Rectangle":
        """
        Instantiates a rectangle from its (x, y) position and width and height.

        :param x:   The x-position of the left side of the rectangle.
        :param y:   The y-position of the top side of the rectangle.
        :param w:   The rectangle's width.
        :param h:   The rectangle's height.
        :return:    The rectangle.
        """
        if w < 0:
            x += w
            w = abs(w)

        if h < 0:
            y += h
            h = abs(h)

        return cls(Point(x, y), Point(x + w - 1, y + h - 1))

    def intersects(self, other: "Rectangle") -> bool:
        """
        Checks if this rectangle intersects with another rectangle.

        :param other:   The other rectangle.
        :return:        True if they intersect.
        """
        # Check for overlap in the x-dimension
        if not self.contains_x(other.left()) and not other.contains_x(self.left()):
            return False

        # Check for overlap in the y-dimension
        if not self.contains_y(other.top()) and not other.contains_y(self.top()):
            return False

        return True

    def intersection(self, other: "Rectangle") -> Optional["Rectangle"]:
        """
        Gets the intersection of this rectangle and another.

        :param other:   The rectangle to intersect with.
        :return:        The rectangle representing the intersection,
                        or None if the two rectangle don't intersect.
        """
        # Check for overlap
        if not self.intersects(other):
            return None

        return Rectangle(Point(max(self.left(), other.left()), max(self.top(), other.top())),
                         Point(min(self.right(), other.right()), min(self.bottom(), other.bottom())))

    def contains_x(self, x: int) -> bool:
        """
        Whether the given x-coordinate is inside the range covered
        by the rectangle.

        :param x:   The x-coordinate.
        :return:    True if the x-coordinate overlaps the rectangle.
        """
        return self.left() <= x <= self.right()

    def contains_y(self, y: int) -> bool:
        """
        Whether the given y-coordinate is inside the range covered
        by the rectangle.

        :param y:   The y-coordinate.
        :return:    True if the y-coordinate overlaps the rectangle.
        """
        return self.top() <= y <= self.bottom()

    def __contains__(self, point: Point) -> bool:
        """
        Whether the given point is inside the rectangular bounds.

        :param point:   The point.
        :return:        True if the point is in the rectangle.
        """
        return self.contains_x(point.x) and self.contains_y(point.y)

    def off_diagonal_points(self) -> Tuple[Point, Point]:
        """
        Gets the corner points that are off the main diagonal
        of this rectangle.

        :return:    The two off-diagonal points.
        """
        return Point(self.p1.x, self.p2.y), Point(self.p2.x, self.p1.y)

    def __hash__(self):
        return hash((self.left(), self.top(), self.right(), self.bottom()))

    def __eq__(self, other):
        return (isinstance(other, Rectangle) and
                self.left() == other.left() and
                self.right() == other.right() and
                self.top() == other.top() and
                self.bottom() == other.bottom())



from abc import abstractmethod
from typing import IO, TypeVar

from ._FileIOBase import FileIOBase, ObjectType, DiskType

SelfType = TypeVar("SelfType", bound="FileWriter")


class FileWriter(FileIOBase[DiskType, ObjectType, SelfType]):
    """
    Base class for classes which write objects to disk.
    """
    def dump(self, obj: ObjectType):
        """
        Serialises an object to the file.

        :param obj:     The object to serialise.
        """
        # If writing fails, ensure the file position remains unchanged
        pre_dump_position = self._file.tell()
        try:
            return self._dump(obj, self._file)
        except Exception:
            self._file.seek(pre_dump_position)
            self._file.truncate()
            raise

    def dumpf(self, filename: str, *objects: ObjectType, append: bool = False):
        """
        Dumps the given objects to a disk file.

        :param filename:    The file to write to.
        :param objects:     The objects to write.
        :param append:      Whether to append the objects.
        """
        with self.set_file(self.open(filename, append)) as writer:
            for obj in objects:
                writer.dump(obj)

    def dumps(self, *objects: ObjectType) -> DiskType:
        """
        Serialises objects to a string.

        :param objects:     The objects to serialise.
        :return:            A string/bytes.
        """
        # Create a new memory file
        memory_file: IO[DiskType] = self.get_memory_file()

        # Write the objects to the memory file
        with self.set_file(memory_file) as writer:
            for obj in objects:
                writer.dump(obj)

        # Return the memory file contents
        memory_file.seek(0)
        return memory_file.read()

    @classmethod
    def open(cls, filename: str, append: bool = False) -> IO[DiskType]:
        """
        Opens the given file for writing using this writer type.

        :param filename:    The file to open.
        :param append:      Whether to append to the end of the file.
        :return:            The file writer.
        """
        return cls._open(filename, 'a' if append else 'w')

    @abstractmethod
    def _dump(self, obj: ObjectType, file: IO[DiskType]):
        """
        Implementation to write the given object to the given file.
        Assumes the file is valid (writable).

        :param obj:     The object to write.
        :param file:    The file to write to.
        """
        pass

    @classmethod
    def _check_file(cls, file: IO[DiskType]):
        if not file.writable():
            raise ValueError("File not writable")



from typing import Type, Union
from argparse import Namespace

from ..meta import instanceoptionalmethod, get_import_code
from ..meta.constants import TRIPLE_QUOTES
from ..meta.typing import VAR_ARGS_TYPE
from ._CLIInstantiable import CLIInstantiable
from ._OptionHandler import OptionHandler
from ._typing import OptionsList


class CLIFactory(OptionHandler):
    """
    Base class for factory classes which instantiate objects
    from command-line arguments.
    """
    @instanceoptionalmethod
    def production_class(self, namespace: Namespace) -> Type:
        """
        Gets the type of object to produce.

        :param namespace:   The namespace from parsing the command-line options.
        :return:            The produced object type.
        """
        raise NotImplementedError(CLIFactory.production_class.__qualname__)

    @instanceoptionalmethod
    def init_args(self, namespace: Namespace) -> VAR_ARGS_TYPE:
        """
        Translates the namespace created by the configured parser
        into arguments to the __init__ method of the production class.

        :param namespace:   The namespace from parsing the command-line options.
        :return:            The arguments to the class.
        """
        raise NotImplementedError(CLIFactory.init_args.__qualname__)

    @instanceoptionalmethod
    def instantiate(self, namespace: Union[Namespace, OptionsList, None] = None) -> CLIInstantiable:
        """
        Produces an instance of the production class from the given namespace.

        :return:    The instance.
        """
        # If values passes unparsed, parse them
        namespace = self._ensure_namespace(namespace)

        # Get the production class
        cls = self.production_class(namespace)

        # Get the init args
        args, kwargs = self.init_args(namespace)

        return cls(*args, **kwargs)

    @staticmethod
    def for_class(cls: Type[CLIInstantiable]) -> Type['CLIFactory']:
        """
        Creates a class which is a factory for the given CLI-instantiable type.

        :param cls:     Any CLI-instantiable type.
        :return:        The factory class.
        """
        # Get the factory class code
        code = CLIFactory.code_for_class(cls)

        # Create a scope dictionary
        scope = {}

        # Execute the class definition code
        exec(code, scope)

        # Return the defined class from the local scope
        return scope[f"{cls.__name__}CLIFactory"]

    @staticmethod
    def code_for_class(cls: Type[CLIInstantiable]) -> str:
        """
        Gets the code to execute to produce a factory for the given
        type of instantiable.

        :param cls:                 The class to produce factory code for.
        :return:                    The factory class definition code.
        """
        # Create the required formatted strings
        cls_name: str = cls.__name__
        cls_qualname: str = cls.__qualname__

        # Get the import code for the types used by CLIFactory
        types_import_code = (
            f"from argparse import Namespace\n"
            f"from typing import Type\n"
            f"\n"
            f"from wai.common.cli import CLIFactory, CLIInstantiable\n"
            f"from wai.common.meta.typing import VAR_ARGS_TYPE\n"
        )

        # TODO: Move to global
        from ..meta.code_repr import code_repr, combine_import_dicts, get_import_dict, get_code

        # Get the code representation for the options used by the class
        option_code_reprs = {option.name: code_repr(option) for option in cls._get_all_options()}

        # Combine the import dicts
        combined_import_dict = combine_import_dicts(*map(get_import_dict, option_code_reprs.values()))

        # TODO: Make sure none of the base imports will be overridden

        # Format the imports
        option_types_import_code: str = "\n".join(import_code for import_code in combined_import_dict.values()) + "\n\n"

        # Get the import code for the class itself
        # (this will be a local import, hence the indentation)
        cls_import_code: str = get_import_code(cls, 8, False)

        # Format the options
        option_definitions = "\n    ".join(f"{name} = {get_code(option_repr)}"
                                           for name, option_repr in option_code_reprs.items())
        if option_definitions == "":
            option_definitions = "# None"

        return (
            f"{types_import_code}\n"
            f"{option_types_import_code}\n"
            f"class {cls_name}CLIFactory(CLIFactory):\n"
            f"    {TRIPLE_QUOTES}\n"
            f"    Factory which instantiates the {cls_name} class.\n"
            f"    {TRIPLE_QUOTES}\n"
            f"    # Options\n"
            f"    {option_definitions}\n"
            f"\n"
            f"    @classmethod\n"
            f"    def production_class(self, namespace: Namespace) -> Type[CLIInstantiable]:\n"
            f"{cls_import_code}\n"
            f"        return {cls_qualname}\n"
            f"\n"
            f"    @classmethod\n"
            f"    def init_args(self, namespace: Namespace) -> VAR_ARGS_TYPE:\n"
            f"        return (namespace,), dict()\n"
        )



from typing import Type, List, Optional, Tuple


class ExceptionChainer:
    """
    Context-manager class which captures exceptions for later use. Designed for
    use in loops where an exception may occur on each iteration, so they can be
    processed at the end.
    """
    def __init__(self, *exc_types: Type[Exception]):
        self.__types: Tuple[Type[Exception]] = tuple()
        self.__exceptions: List[Optional[Exception]] = []

        self.set_types(*exc_types)

    def exception_count(self) -> int:
        """
        Gets the number of exceptions captured.
        """
        return sum(1 for e in self.__exceptions if e is not None)

    def no_exception_count(self) -> int:
        """
        Gets the number of iterations that occurred with no exception.
        """
        return len(self) - self.exception_count()

    def reset(self):
        """
        Clears the list of captured exceptions.
        """
        self.__exceptions.clear()

    def set_types(self, *exc_types: Type[Exception]):
        """
        Sets the types of exceptions this chainer will capture.
        If no types are given, captures all exceptions.

        :param exc_types:   The types of exception.
        """
        self.__types = exc_types if len(exc_types) > 0 else (Exception,)

    def exceptions(self, remove_missing: bool = False) -> List[Optional[Exception]]:
        """
        Gets the list of exceptions in captured order.

        :param remove_missing:  Whether to remove None values (iterations
                                that didn't throw).
        :return:                The list of exceptions.
        """
        return [e for e in self.__exceptions if e is not None or not remove_missing]

    def raise_last(self):
        """
        Raises the last exception captured. If there isn't one,
        simply returns.
        """
        # Get the exceptions
        exceptions = self.exceptions(True)

        # If there are any, raise the last one
        if len(exceptions) > 0:
            raise exceptions[-1]

    def raise_chained(self, reversed: bool = False):
        """
        Raises the captured exceptions. Each exception is treated as
        the cause of the following exception (or vice versa if reversed
        is set). If not exceptions have been captured, simply returns.

        :param reversed:    Whether to chain in reverse order.
        """
        # Get the exceptions
        exceptions = self.exceptions(True)

        # Just return if there are no exceptions
        if len(exceptions) == 0:
            return

        # Reverse if requested
        if reversed:
            exceptions.reverse()

        # Chain the exceptions
        exception = exceptions[0]
        for i in range(1, len(exceptions)):
            try:
                raise exceptions[i] from exception
            except Exception as e:
                exception = e

        # Raise the chained exception
        raise exception

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            # No exception occurred
            self.__exceptions.append(None)
            return True
        elif issubclass(exc_type, self.__types):
            # An exception we are looking for occurred
            self.__exceptions.append(exc_val)
            return True
        else:
            # An unexpected exception occurred, so don't suppress it
            return False

    def __len__(self) -> int:
        return len(self.__exceptions)



from argparse import Namespace
from typing import Union, Any

from .options import Option
from ._OptionHandler import OptionHandler
from ._typing import OptionsList


class OptionValueHandler(OptionHandler):
    """
    Base class for classes which handle options and their values.
    """
    @property
    def namespace(self) -> Namespace:
        """
        Gets the namespace that holds this handler's values.

        :return:    The namespace.
        """
        raise NotImplementedError("OptionValueHandler.namespace")

    def get_option_value(self, option: Union[str, Option]) -> Any:
        """
        Gets the current value for an option.

        :param option:  The option, by name or reference.
        :return:        The option's value.
        """
        # Convert option names to their descriptors
        if isinstance(option, str):
            option = self._get_option(option)

        # Get the value from the namespace by name
        return option.__get__(self, type(self))

    def set_option_value(self, option: Union[str, Option], value: Any):
        """
        Sets the current value for an option.

        :param option:  The option, by name or reference.
        :param value:   The value to set.
        """
        # Convert option names to their descriptors
        if isinstance(option, str):
            option = self._get_option(option)

        # Get the value from the namespace by name
        return option.__set__(self, value)

    def to_options_list(self) -> OptionsList:
        """
        Gets the values of all options as an options-list.
        """
        options_list = []

        for option in self._get_all_options():
            options_list += option.get_as_options_list(self)

        return options_list



import os

from ._util import ensure_path


class PathContextManager:
    """
    Context manager class which changes the cwd to a new
    path. If a relative path is given, it is considered
    relative to the current cwd.
    """
    def __init__(self, path: str, ensure: bool = False):
        self.cwd: str = ""  # The cwd before changing to the given path; gets set on entry
        self.path: str = path  # The path to change to
        self.ensure: bool = ensure  # Whether to ensure the path exists by creating it if needed

    def __enter__(self):
        # Save the cwd
        self.cwd = os.path.abspath(os.getcwd())

        # Optionally ensure the path exists
        if self.ensure:
            ensure_path(self.path)

        # Calculate the new path
        new_path = os.path.join(self.cwd, self.path) if not os.path.isabs(self.path) else self.path

        # Change to the new path
        os.chdir(new_path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Change back to the previous cwd
        os.chdir(self.cwd)

        # Re-raise any exceptions
        return False


# Sugar definition
offset_cwd = PathContextManager



from argparse import REMAINDER
from typing import Union, Any, Type, Iterable, Tuple, TypeVar, List

from ...meta import non_default_kwargs
from ...meta.code_repr import CodeRepresentation, from_init
from .._CLIRepresentable import CLIRepresentable, cli_repr, from_cli_repr, is_cli_representable_type
from .._typing import OptionsList
from ._Option import Option

# The type of class the option takes
ClassType = TypeVar("ClassType", bound=CLIRepresentable)

# The set of actions allowed with this type of option
ALLOWED_ACTIONS = {"store", "append"}


class TypedOption(Option):
    """
    An option which takes a specified type of CLI-representable objects as values.
    """
    def __init__(self,
                 *flags: str,
                 type: Type[ClassType],
                 action: str = ...,
                 nargs: Union[int, str] = ...,
                 choices: Iterable[Union[str, ClassType]] = ...,
                 default: Union[ClassType, List[ClassType], None] = ...,
                 required: bool = ...,
                 help: str = ...,
                 metavar: Union[str, Tuple[str, ...]] = ...):
        # Capture the code-representation of the option
        code_representation: CodeRepresentation = from_init(self, locals())

        # Check the type is CLI-representable
        if not is_cli_representable_type(type):
            raise TypeError(f"Type {type.__qualname__} is not a CLI-representable type")
        self._type = type

        # Check the action is allowed
        if action is not ... and action not in ALLOWED_ACTIONS:
            raise ValueError(f"Action must be one of {ALLOWED_ACTIONS}")
        self._action = action if action is not ... else "store"

        # Check the number of arguments is allowed
        if nargs == "?" or nargs == REMAINDER:
            raise ValueError(f"Disallowed nargs value {nargs}")
        self._nargs = nargs if nargs is not ... else None

        # Save our optional status
        self._optional = not required if required is not ... else True

        # See if our values should be list-valued or not
        self._is_list_valued = action == "append" or nargs is not ...

        # Format choices as strings
        if choices is not ...:
            formatted_choices = []
            for choice in choices:
                # Convert CLI representations into actual objects
                if isinstance(choice, str):
                    choice = from_cli_repr(type, choice)

                # Make sure they are instances of the defined type
                if not isinstance(choice, type):
                    raise TypeError(f"Choice is not of the declared type")

                formatted_choices.append(cli_repr(choice))
            choices = formatted_choices

        # Format default value
        if default is ... and self._optional:
            default = [] if self._is_list_valued else None

        # Extract the set of keyword arguments
        kwargs = non_default_kwargs(TypedOption.__init__, locals())

        super().__init__(code_representation, *flags, **kwargs)

    @property
    def type(self) -> Type[ClassType]:
        """
        Gets the type of value this option takes.
        """
        return self._type

    def _namespace_value_to_internal_value(self, namespace_value: Any) -> Any:
        if isinstance(namespace_value, list):
            return [from_cli_repr(self._type, cli_string) for cli_string in namespace_value]
        elif isinstance(namespace_value, str):
            return from_cli_repr(self._type, namespace_value)
        else:
            return None

    def _internal_value_to_namespace_value(self, internal_value: Any) -> Any:
        if isinstance(internal_value, list):
            return [cli_repr(value) for value in internal_value]
        elif isinstance(internal_value, self._type):
            return cli_repr(internal_value)
        else:
            return None

    def _namespace_value_to_options_list(self, namespace_value: Any) -> OptionsList:
        if isinstance(namespace_value, list):
            length = len(namespace_value)
            nargs = self._nargs if self._nargs is not None else 1
            options_list = []
            if self._optional and length == 0:
                pass
            if self._action == "store" or isinstance(nargs, str):
                options_list.append(self._flags[0])
                for val in namespace_value:
                    options_list.append(cli_repr(val))
            else:
                for i in range(0, len(namespace_value), nargs):
                    options_list.append(self._flags[0])
                    for val in namespace_value[i:i + nargs]:
                        options_list.append(cli_repr(val))
            return options_list
        else:
            if self._optional and namespace_value is None:
                return []
            return [self._flags[0], namespace_value]

    def _validate_namespace_value(self, namespace_value: Any):
        if self._is_list_valued:
            if not isinstance(namespace_value, list):
                raise TypeError(f"Class option expects namespace values as lists")

            if not all(isinstance(element, str) for element in namespace_value):
                raise TypeError(f"Class option expects all elements of namespace list to be strings")
        elif not self._optional or namespace_value is not None:
            if not isinstance(namespace_value, str):
                raise TypeError(f"Class option expects namespace values as strings")

    def _validate_internal_value(self, internal_value: Any):
        if self._is_list_valued:
            self._validate_internal_value_list(internal_value)
        else:
            self._validate_internal_value_singular(internal_value)

    def _validate_internal_value_singular(self, value: Any):
        """
        Validate's the given value when trying to set the value
        of this option. Assumes the value should be a singular value.

        :param value:       The value attempting to be set.
        :return:            The actual value that should be set.
        :raises Exception:  If the value cannot be validated.
        """
        # Make sure the value is of our type (or None if we're optional)
        if not self._optional or value is not None:
            if not isinstance(value, self._type):
                raise TypeError(f"Expected {self._type.__name__} but got {type(value).__name__}")

    def _validate_internal_value_list(self, value: Any):
        """
        Validate's the given value when trying to set the value
        of this option. Assumes the value should be a list of values.

        :param value:       The value attempting to be set.
        :return:            The actual value that should be set.
        :raises Exception:  If the value cannot be validated.
        """
        # Make sure the value is a list
        if not isinstance(value, list):
            raise TypeError(f"Value must be a list")

        # Make sure the list has the correct number of values in it
        length = len(value)
        if self._optional and length == 0:
            pass
        elif isinstance(self._nargs, int):
            if self._action == "store":
                if length != self._nargs:
                    raise ValueError(f"Requires list of length {self._nargs}")
            else:
                if length % self._nargs != 0:
                    raise ValueError(f"Requires list of length a multiple of {self._nargs}")
        elif self._nargs == "+":
            if length == 0:
                raise ValueError(f"Requires at least one item")

        if any(not isinstance(element, self._type) for element in value):
            raise TypeError(f"All values must be of the declared type {self._type.__name__}")



from functools import wraps
from typing import Type

from ..meta import all_as_kwargs
from ..meta.typing import GenericDecorator, GenericCallable


def ensure_error_type(error_type: Type[Exception],
                      format_message: str = "Unhandled exception of type {0.__class__.__name__}: "
                                            "{0}") -> GenericDecorator:
    """
    Decorator which ensures any exceptions coming from the called function
    are of the given error-type. Any exceptions that the function raises
    that are already of the correct type (or a sub-type) are passed through
    normally, any other errors are chained with an error of the given type.

    :param error_type:      The type of error to ensure.
    :param format_message:  The message to format if an exception of another type occurs.
                            Can include format strings involving any arguments passed to
                            the function (by name only), and the positional argument 0
                            which is the exception that occurred.
    :return:                The decorator function.
    """
    def decorator(function: GenericCallable) -> GenericCallable:
        @wraps(function)
        def with_ensured_error_type(*args, **kwargs):
            try:
                # Call the function
                return function(*args, **kwargs)
            except error_type:
                # Re-raise any exception of the requested error-type
                raise
            except Exception as e:
                # Wrap any other exception in the requested error-type
                all_kwargs = all_as_kwargs(function, *args, **kwargs)
                raise error_type(format_message.format(e, **all_kwargs)) from e

        return with_ensured_error_type

    return decorator



from argparse import ArgumentParser, Namespace
from typing import Type, Iterator, Union, Optional

from .options import Option
from ._ArgumentParserConfigurer import ArgumentParserConfigurer
from ._typing import OptionsList, is_options_list


class OptionHandler(ArgumentParserConfigurer):
    """
    Super class for option-handling classes.
    """
    @classmethod
    def configure_parser(cls, parser: ArgumentParser):
        # Configure the parser with each option
        for option in cls._get_all_options():
            option.configure_parser(parser)

    @classmethod
    def get_all_options(cls) -> Iterator[str]:
        """
        Gets the names of all options on this class.

        :return:    An iterator of option names.
        """
        return (option.name for option in cls._get_all_options())

    @classmethod
    def get_common_options(cls, other: Type['OptionHandler']) -> Iterator[str]:
        """
        Gets the options common to this handler and another.

        :param other:   The other option handler.
        :return:        The set of common option names.
        """
        # Get the other's options, keyed by name
        other_options = {option.name: option for option in other._get_all_options()}

        # Return all option names which refer to identical options
        return (option.name
                for option in cls._get_all_options()
                if option.name in other_options
                and other_options[option.name] == option)

    @classmethod
    def get_help_text_for_option(cls, option: Option) -> Optional[str]:
        """
        Allows the help text for options to be set by the owning
        class.

        :param option:  The option to get the help text for.
        :return:        The help text for the option,
                        or None if not provided.
        """
        return None

    @classmethod
    def _get_all_options(cls) -> Iterator[Option]:
        """
        Gets all options for this option handler.

        :return:    An iterator over the options.
        """
        for name in dir(cls):
            # Option names can't start with underscores
            if name.startswith("_"):
                continue

            # Get the attribute
            obj = getattr(cls, name)

            # If it's an option descriptor, yield it
            if isinstance(obj, Option):
                yield obj

    @classmethod
    def _get_option(cls, name: str) -> Option:
        """
        Gets an option by name.

        :param name:    The option's name.
        :return:        The option.
        """
        # Get the attribute
        attr = getattr(cls, name, None)

        # Make sure it is an option
        if not isinstance(attr, Option):
            raise NameError(f"Class '{cls.__qualname__} has no option '{name}'")

        return attr

    @classmethod
    def _ensure_namespace(cls, namespace: Union[Namespace, OptionsList, None] = None) -> Namespace:
        """
        Helper method which takes a namespace or an options list
        and returns a namespace. If None is provided, uses
        sys.argv.

        :param namespace:   The namespace or options list.
        :return:            The namespace.
        """
        # If already a namespace, return it
        if isinstance(namespace, Namespace):
            return namespace

        # Must be an options list if not a namespace
        if namespace is not None and not is_options_list(namespace):
            raise ValueError(f"{namespace} is not a Namespace or an options list")

        # Parse the options list (uses sys.argv if namespace is None)
        return cls.get_configured_parser().parse_args(namespace)



from typing import IO

from ._StringSerialiser import StringSerialiser, NULL_BYTE


class NullTerminatedStringSerialiser(StringSerialiser):
    """
    Serialises a string by appending a null byte after the
    encoded string.
    """
    def __init__(self,
                 encoding: str = "utf-8"):
        super().__init__(encoding)

    def _check_encoded(self, encoded: bytes):
        # Check the encoded string doesn't already contain null-bytes
        if NULL_BYTE in encoded:
            raise ValueError(f"Encoded string {encoded} contains a null-byte")

    def _serialise_encoded(self, encoded: bytes, stream: IO[bytes]):
        stream.write(encoded)
        stream.write(NULL_BYTE)

    def _deserialise_encoded(self, stream: IO[bytes]) -> bytes:
        # Read the bytes
        byte_string: bytes = b''
        while not byte_string.endswith(NULL_BYTE):
            byte_string += stream.read(1)

        # Return without the null-byte
        return byte_string[:-1]



import io
from typing import List, Optional, Union, Type

import csv

# The type of the header section
HEADER_TYPE = List[str]

# The type of the column types
TYPES_TYPE = List[Union[Type[int], Type[float], Type[str]]]

# The type of the data section
VALUE_TYPE = Union[int, float, str]
ROW_TYPE = List[VALUE_TYPE]
DATA_TYPE = List[ROW_TYPE]


class CSVFile:
    """
    Class representing a CSV file loaded into memory.
    """
    def __init__(self,
                 header: Optional[HEADER_TYPE] = None,
                 types: Optional[TYPES_TYPE] = None,
                 data: Optional[DATA_TYPE] = None,
                 dialect: Optional[csv.Dialect] = None):
        # Create an empty dataset if no data is given
        if data is None:
            data = []

        # If no types are given, assume all are string
        if types is None:
            types = [str] * len(data[0]) if len(data) > 0 else []

        self.header: Optional[HEADER_TYPE] = header
        self.types: TYPES_TYPE = types
        self.data: DATA_TYPE = data
        self.dialect: Optional[csv.Dialect] = dialect

    def get_column(self, column_index):
        """
        Gets a column of data.

        :param column_index:    The column to copy.
        :return:                The column of data (as a list).
        """
        return [row[column_index] for row in self.data]

    def __str__(self) -> str:
        # Create a string buffer
        buffer: io.StringIO = io.StringIO()

        # Create a CSV writer to write to it
        writer: csv.writer = csv.writer(buffer, self.dialect)

        # Write the header
        writer.writerow(self.header)

        # Write the data rows
        for row in self.data:
            writer.writerow(row)

        # Return the buffered string
        buffer.seek(0)
        return buffer.read()



from argparse import Namespace
from typing import Union, Any

from .util import to_cli_string, from_cli_string
from ._CLIRepresentable import CLIRepresentable, cli_repr, from_cli_repr
from ._OptionValueHandler import OptionValueHandler
from ._typing import OptionsList


class CLIInstantiable(CLIRepresentable, OptionValueHandler):
    """
    Base class for types that can be instantiated from the
    command-line.
    """
    def __init__(self,
                 _namespace: Union[Namespace, OptionsList, None] = None,
                 **internal: Any):
        self._namespace: Namespace = (
            self._ensure_namespace(_namespace)
            if _namespace is not None or len(internal) == 0
            else Namespace()
        )

        for option in self._get_all_options():
            # If an option's value is supplied in internal format, set it
            if option.name in internal:
                self.set_option_value(option, internal[option.name])

            # If the option has not been set, set it to its default
            elif not hasattr(self._namespace, option.name):
                option.set_from_options_list(self, [])

    @property
    def namespace(self) -> Namespace:
        return self._namespace

    def cli_repr(self) -> str:
        return to_cli_string([cli_repr(type(self))] + self.to_options_list())

    @classmethod
    def from_cli_repr(cls, cli_string: str) -> 'CLIInstantiable':
        # Split into class and option parts
        options_list = from_cli_string(cli_string)

        # Parse the class
        klass = from_cli_repr(type, options_list[0])

        # Make sure it is a sub-class of our type
        if not issubclass(klass, cls):
            raise TypeError(f"{options_list[0]} is not a sub-class of {cli_repr(cls)}")

        options_list = options_list[1:]

        return klass(options_list)



"""
Module for common functionality for get_argument_to_typevar between Python versions.
"""
from typing import Type, TypeVar, Tuple

import typing_inspect


def check_args(cls: Type, generic_base_class: Type, typevar: TypeVar) -> int:
    """
    Performs argument-validation for get_argument_to_typevar.

    :param cls:                 The sub-class specifying the argument.
    :param generic_base_class:  The generic base-class specifying the type variable.
    :param typevar:             The type variable to get the argument for.
    :return:                    The index of the typevar in the base class' parameters.
    """
    # Make sure the class derives from the base-class
    if not issubclass(cls, generic_base_class):
        raise ValueError(f"{cls.__name__} does not derive from {generic_base_class.__name__}")

    # Make sure the base class is generic
    if not typing_inspect.is_generic_type(generic_base_class):
        raise TypeError(f"{generic_base_class.__name__} is not a generic type")

    # Get the type parameters to the generic base class
    parameters = typing_inspect.get_parameters(generic_base_class)

    # Make sure the type variable is a parameter to the base class
    if typevar not in parameters:
        raise ValueError(f"{typevar} is not a generic parameter of {generic_base_class.__name__}")

    return parameters.index(typevar)



from typing import Generic, TypeVar
from weakref import WeakKeyDictionary

ValueType = TypeVar("ValueType")


class InstanceProperty(Generic[ValueType]):
    """
    A basic property which simply holds a value for each instance object. As
    instances of classes which use this property are used as the key to a dictionary
    holding the value, the instances must be hashable, and the hash must not rely on
    the property for its hash calculation.
    """
    def __init__(self, default_value: ValueType = None):
        self._cache = WeakKeyDictionary()
        self._default: ValueType = default_value

    def __get__(self, instance, owner) -> ValueType:
        if instance in self._cache:
            return self._cache[instance]

        return self._default

    def __set__(self, instance, value: ValueType):
        self._cache[instance] = value

    def __delete__(self, instance):
        del self._cache[instance]



"""
Utility functions for working with code representations.
"""
from inspect import signature, Signature, Parameter

from ._error import ConflictingImports
from ._functional import code_repr
from ._typing import CodeRepresentation, ImportDict


def get_import_dict(code_representation: CodeRepresentation) -> ImportDict:
    """
    Gets the import dictionary from a code representation.

    :param code_representation:     The code representation.
    :return:                        The import dictionary.
    """
    return code_representation[0]


def get_code(code_representation: CodeRepresentation) -> str:
    """
    Gets the code from a code representation.

    :param code_representation:     The code representation.
    :return:                        The code.
    """
    return code_representation[1]


def combine_import_dicts(*import_dicts: ImportDict) -> ImportDict:
    """
    Combines multiple import dictionaries into one, ensuring no
    imported names are lost.

    :param import_dicts:            The import dictionaries to combine.
    :return:                        The combined import dictionary.
    :raises ConflictingImports:     If there are conflicting imports between
                                    the import dictionaries.
    """
    result = {}

    for import_dict in import_dicts:
        for imported_name, import_code in import_dict.items():
            if imported_name in result:
                if import_code != result[imported_name]:
                    raise ConflictingImports(imported_name, import_code, result[imported_name])
            else:
                result[imported_name] = import_code

    return result


def from_init(self, locals_) -> CodeRepresentation:
    """
    Utility for when the code-representation is completely
    determined in the __init__ method by the arguments passed
    to that method.

    :param self:        The object being initialised.
    :param locals_:     The local scope of the __init__ method.
    :return:            The code-representation for the object.
    """
    # Get the type being initialised
    cls = type(self)

    # Initialise the import dict with the import for the class itself
    import_dict = get_import_dict(code_repr(cls))

    # Get the method signature of the __init__ method
    init_signature: Signature = signature(self.__init__)

    # Start the representation
    code = f"{cls.__qualname__}("

    # Format each argument in turn
    first = True
    for name, parameter in init_signature.parameters.items():
        parameter_string = ""
        value = locals_[name]
        if parameter.kind in {Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY}:
            if value is not parameter.default:
                value_repr = code_repr(value)
                import_dict = combine_import_dicts(import_dict, get_import_dict(value_repr))
                parameter_string = (f"{name}={get_code(value_repr)}"
                                    if parameter.kind == Parameter.KEYWORD_ONLY
                                    else get_code(value_repr))
        elif parameter.kind == Parameter.VAR_POSITIONAL:
            if len(value) > 0:
                value_reprs = tuple(map(code_repr, value))
                import_dict = combine_import_dicts(import_dict, *map(get_import_dict, value_reprs))
                parameter_string = ", ".join(map(get_code, value_reprs))
        elif parameter.kind == Parameter.VAR_KEYWORD:
            if len(value) > 0:
                value_reprs = {key: code_repr(value) for key, value in value.items()}
                import_dict = combine_import_dicts(import_dict, *map(get_import_dict, value_reprs.values()))
                parameter_string = ", ".join(f"{key}={get_code(value_repr)}" for key, value_repr in value_reprs.items())

        if parameter_string != "":
            if not first:
                code += ", "
            else:
                first = False
            code += parameter_string

    code += ")"

    return import_dict, code



from typing import TypeVar, Callable

# The type of the dynamic default values
DefaultType = TypeVar("DefaultType")


class _DynamicDefault:
    """
    Class which determines that the default value for a
    function should be recalculated per-call.
    """
    def __init__(self, default: Callable[[], DefaultType]):
        self._default: Callable[[], DefaultType] = default

    def get(self) -> DefaultType:
        return self._default()


dynamic_default = _DynamicDefault



from typing import Tuple, Any, Dict, Callable, TypeVar

# The type of a standard *args parameter
POSITIONAL_ARGS_TYPE = Tuple[Any, ...]

# The type of a standard **kwargs parameter
KEYWORD_ARGS_TYPE = Dict[str, Any]

# The type of *args, **kwargs combined
VAR_ARGS_TYPE = Tuple[POSITIONAL_ARGS_TYPE, KEYWORD_ARGS_TYPE]

# The type of a callable (non-generic)
AnyCallable = Callable[[Any], Any]

# The generic type of arguments to a callable
ArgType = TypeVar("ArgType")

# The generic return type of a callable
ReturnType = TypeVar("ReturnType")

# The type of a generic callable
GenericCallable = Callable[[ArgType], ReturnType]

# The type of a generic decorator function
GenericDecorator = Callable[[GenericCallable], GenericCallable]



from typing import IO

from ._StringSerialiser import StringSerialiser, NULL_BYTE


class FixedLengthStringSerialiser(StringSerialiser):
    """
    Serialises strings to a fixed length. If the string
    is shorter than the fixed length, it is padded with
    null-bytes. If the string is longer than the fixed
    length, it will be truncated, or raise an exception
    if selected.

    N.B. Be careful using this serialiser with variable
    length encoding schemes, as partial letters may be
    truncated.
    """
    def __init__(self,
                 length: int,
                 encoding: str = "utf-8",
                 ensure_null_terminated: bool = False,
                 raise_on_truncate: bool = False):
        super().__init__(encoding)

        # Make sure length is positive
        if length <= 0:
            raise ValueError(f"Length must be a positive value, not {length}")

        self._length: int = length - 1 if ensure_null_terminated else length
        self._ensure_null_terminated: bool = ensure_null_terminated
        self._raise_on_truncate: bool = raise_on_truncate

    def _check_encoded(self, encoded: bytes):
        # Check for truncation
        if self._raise_on_truncate:
            if len(encoded) > self._length:
                raise ValueError(f"Encoded string of length {len(encoded)} will truncate")

    def _serialise_encoded(self, encoded: bytes, stream: IO[bytes]):
        # Write the (possibly truncated) string bytes
        stream.write(encoded[:self._length])

        # Pad to length with null-bytes as necessary
        if len(encoded) < self._length:
            stream.write(NULL_BYTE * (self._length - len(encoded)))

        # Write the null terminator if selected
        if self._ensure_null_terminated:
            stream.write(NULL_BYTE)

    def _deserialise_encoded(self, stream: IO[bytes]) -> bytes:
        # Read the encoded string
        encoded = stream.read(self._length)

        # Remove any padding
        while encoded.endswith(NULL_BYTE):
            encoded = encoded[:-1]

        # Consume the null-terminator if expected
        if self._ensure_null_terminated:
            stream.read(1)

        return encoded



from typing import IO

from .._FileWriter import FileWriter
from ._ARFFFile import ARFFFile


class ARFFFileWriter(FileWriter[str, ARFFFile, "ARFFFileWriter"]):
    """
    File writer for ARFF files.
    """
    def _dump(self, obj: ARFFFile, file: IO[str]):
        file.write(str(obj))



import functools
from inspect import Signature, signature, BoundArguments
from typing import TypeVar, Callable, Any

from ._dynamic_default import dynamic_default

# The return type of the wrapped function
ReturnType = TypeVar("ReturnType")


def with_dynamic_defaults(function: Callable[[Any], ReturnType]):
    """
    Decorator which processes dynamic defaults for the wrapped function.

    :param function:    The wrapped function with dynamic defaults.
    :return:            The decorated function.
    """
    # Get the signature of the wrapped function
    function_signature: Signature = signature(function)

    @functools.wraps(function)
    def resolve_dynamic_defaults(*args, **kwargs):
        # Bind the arguments
        binding: BoundArguments = function_signature.bind(*args, **kwargs)
        binding.apply_defaults()

        # Resolve any dynamic defaults
        for name in function_signature.parameters:
            if isinstance(binding.arguments[name], dynamic_default):
                binding.arguments[name] = binding.arguments[name].get()

        # Call the function with the resolved arguments
        return function(*binding.args, **binding.kwargs)

    return resolve_dynamic_defaults



from abc import ABC
from typing import Optional, List, Any

from ._DataType import DataType


class AbstractField(ABC):
    """
    A single report field identifier.
    """
    # The separator
    SEPARATOR: str = "\t"

    # The escaped separator
    SEPARATOR_ESCAPED: str = "\\t"

    # The replacement for tabs in the fields when displaying in the GUI.
    SEPARATOR_DISPLAY: str = " | "

    def __init__(self, *args):
        # Infer arguments
        if len(args) == 0:
            name = None
            datatype = DataType.UNKNOWN
        elif len(args) == 1 and isinstance(args[0], AbstractField):
            name = args[0].name
            datatype = args[0].datatype
        elif len(args) == 2 and isinstance(args[0], (str, type(None))) and isinstance(args[1], (DataType, type(None))):
            name = args[0]
            datatype = args[1]
        elif len(args) == 3 and isinstance(args[0], str) and isinstance(args[1], str) and \
                isinstance(args[2], (DataType, type(None))):
            name = args[0] + AbstractField.SEPARATOR + args[1]
            datatype = args[2]
        else:
            raise ValueError("Bad arguments to initialise AbstractField")

        self.name: Optional[str] = self.fix_string(self.unescape(name)) if name is not None else None
        self.datatype: DataType = datatype if datatype is not None else DataType.UNKNOWN
        self.prefix: Optional[str] = None
        self.suffix: Optional[str] = None

    def is_compound(self) -> bool:
        """
        Checks whether the name is a compound one, i.e., contains a SEPARATOR.

        :return:    True if the name is a compound one.
        """
        return self.name is not None and AbstractField.SEPARATOR in self.name

    def split(self) -> List[str]:
        """
        Returns the name split up into its single parts.

        :return:    The parts of the compound name.
        """
        if self.name is None:
            return []

        return self.name.split(AbstractField.SEPARATOR)

    def to_parseable_string(self) -> str:
        """
        Returns the name and the type (format: name[type]). Can be restored with parseField.

        :return:    The name and type.
        """
        if self.name is None:
            return ""

        return self.escape(self.name, AbstractField.SEPARATOR_ESCAPED) + f"[{self.datatype}]"

    def to_string(self) -> str:
        """
        Returns the name of the field.

        :return:    The name.
        """
        if self.name is None:
            return ""

        return self.escape(self.name, AbstractField.SEPARATOR_ESCAPED)

    def to_display_string(self) -> str:
        """
        Returns the name of the field.

        :return:    The name.
        """
        if self.name is None:
            return ""

        return self.escape(self.name, AbstractField.SEPARATOR_DISPLAY)

    def value_of(self, string: str) -> Any:
        """
        Parse string to appropriate datatype.

        :param string:  The string to parse.
        :return:        The parsed object: STRING, NUMERIC or BOOLEAN.
        """
        try:
            return self.datatype.convert(self.fix_string(string))
        except Exception:
            return None

    @staticmethod
    def escape(name: str, separator: str) -> str:
        """
        Escapes the name.

        :param name:        The name to escape.
        :param separator:   The separator to use.
        :return:            The escaped name.
        """
        return (
            name.replace(AbstractField.SEPARATOR, separator)
                .replace("[", "\\[")
                .replace("]", "\\]")
        )

    @staticmethod
    def unescape(name: str) -> str:
        """
        Unescapes the name.

        :param name:    The name to unescape.
        :return:        The unescaped name.
        """
        return (
            name.replace(AbstractField.SEPARATOR_ESCAPED, AbstractField.SEPARATOR)
                .replace("\\[", "[")
                .replace("\\]", "]")
        )

    def get_clone(self) -> "AbstractField":
        """
        Returns a cloned copy of itself.

        :return:    The clone.
        """
        return self.__class__(self.name, self.datatype)

    def compare_to(self, other: Any) -> int:
        """
        Compares this object with the specified object for order.  Returns a
        negative integer, zero, or a positive integer as this object is less
        than, equal to, or greater than the specified object.

        :param other:   The object to be compared.
        :return:        A negative integer, zero, or a positive integer as this object
                        is less than, equal to, or greater than the specified object.
        """
        if other is None:
            return 1

        if not isinstance(other, AbstractField):
            return -1

        if self.name is None and other.name is None:
            return 0

        if self.name is None:
            return -1

        if other.name is None:
            return 1

        return (-1 if self.name.lower() < other.name.lower() else
                0 if self.name.lower() == other.name.lower() else
                1)

    def equals(self, other: Any) -> bool:
        """
        The equals method (uses only the name).

        :param other:   The object to compare with.
        :return:        True if the same name.
        """
        return self.compare_to(other) == 0

    def hash_code(self) -> int:
        """
        Hashcode so can be used as hashtable key. Returns the hashcode of the
        name.

        :return:    The hashcode.
        """
        if self.name is None:
            return -1

        return hash(self.name)

    def get_prefix(self) -> Optional[str]:
        """
        Returns the prefix for compound fields.

        :return:    The prefix, None if not compound field.
        """
        if not self.is_compound():
            return None

        if self.prefix is None:
            self.prefix = self.split()[0]

        return self.prefix

    def get_suffix(self) -> Optional[str]:
        """
        Returns the suffix for compound fields.

        :return:    The suffix, None if not compound field.
        """
        if not self.is_compound():
            return None

        if self.suffix is None:
            self.suffix = self.split()[1]

        return self.suffix

    @staticmethod
    def fix_string(string: str) -> str:
        """
        Replace ' with `.

        :param string:  The string to process.
        :return:        The processed string.
        """
        return string.replace("'", "`")

    def __str__(self) -> str:
        return self.to_string()

    def __hash__(self) -> int:
        return self.hash_code()

    def __eq__(self, other) -> bool:
        return self.equals(other)



from typing import Any

from ...meta.code_repr import from_init, CodeRepresentation
from .._typing import OptionsList
from ._Option import Option


class FlagOption(Option):
    """
    Option which stores True or False based on if it is provided or not.
    """
    def __init__(self,
                 *flags: str,
                 invert: bool = False,
                 help: str = ...):
        # Capture the code-representation of the option
        code_representation: CodeRepresentation = from_init(self, locals())

        # Save the invert flag
        self._invert: bool = invert

        super().__init__(code_representation,
                         *flags,
                         action="store_false" if invert else "store_true",
                         help=help)

    def _namespace_value_to_internal_value(self, namespace_value: Any) -> bool:
        return namespace_value

    def _internal_value_to_namespace_value(self, internal_value: Any) -> bool:
        return internal_value

    def _namespace_value_to_options_list(self, namespace_value: Any) -> OptionsList:
        # Flag should be present if value is True XOR invert is True
        return [self.flags[0]] if namespace_value != self._invert else []

    def _validate_internal_value(self, value: Any):
        # Make sure the value is boolean
        if not isinstance(value, bool):
            raise TypeError(f"Flag options only take boolean values, not {value}")

    def _validate_namespace_value(self, namespace_value: Any):
        if not isinstance(namespace_value, bool):
            raise ValueError(f"Flag option expects boolean namespace value but got {namespace_value}")



from typing import IO

from .._Serialiser import Serialiser
from ._StringSerialiser import StringSerialiser
from ._IntSerialiser import IntSerialiser


class LengthPrefixedStringSerialiser(StringSerialiser):
    """
    Serialiser of strings. Serialises the length of the string
    before serialising the string itself.
    """
    def __init__(self,
                 encoding: str = "utf-8",
                 length_serialiser: Serialiser[int] = IntSerialiser()):
        super().__init__(encoding)

        self._length_serialiser: Serialiser[int] = length_serialiser

    def _check_encoded(self, encoded: bytes):
        # No checks needed
        pass

    def _serialise_encoded(self, encoded: bytes, stream: IO[bytes]):
        # Get the byte-length of the string
        byte_length: int = len(encoded)

        # Serialise the byte length
        self._length_serialiser.serialise(byte_length, stream)

        # Write the string bytes
        stream.write(encoded)

    def _deserialise_encoded(self, stream: IO[bytes]) -> bytes:
        # Read the length from the stream
        byte_length: int = self._length_serialiser.deserialise(stream)

        return stream.read(byte_length)



import gzip
from io import StringIO, BytesIO
from typing import TypeVar, Generic, Union, Type, Optional, IO, Tuple, Callable

from ..meta.typing import TypeVarProperty

# The type of object that is read/written
ObjectType = TypeVar("ObjectType")

# The type of data the file contains (text/binary)
DiskType = TypeVar("DiskType", bytes, str)

# The type of cls (for self-referential methods)
SelfType = TypeVar("SelfType", bound="FileIOBase")

# The type of the open() function
OpenFunc = Callable[[str, str], IO[DiskType]]


class FileIOBase(Generic[DiskType, ObjectType, SelfType]):
    """
    Base class for FileReader and FileWriter.
    """
    # TypeVar properties
    _disk_type = TypeVarProperty(DiskType)
    _object_type = TypeVarProperty(ObjectType)

    def __init__(self):
        self._file: Optional[IO[DiskType]] = None

    def set_file(self, file: Union[DiskType, IO[DiskType]]) -> SelfType:
        """
        Sets the file for this reader/writer to operate on.

        :param file:    The file.
        :return:        Itself (chainable).
        """
        # Convert raw strings/bytes into in-memory files
        if isinstance(file, self._disk_type):
            file = self.get_memory_file(file)

        # Make sure the file is valid
        self._check_file(file)

        # Set the file
        self._file = file

        return self

    def has_file(self) -> bool:
        """
        Checks if the reader/writer has a file to operate on.
        """
        return self._file is not None

    def ensure_file(self):
        """
        Ensures that the reader/writer has a file to operate on.
        """
        if not self.has_file():
            raise AttributeError("No file set")

    @classmethod
    def get_memory_file(cls, initial_value: Optional[DiskType] = None) -> Union[StringIO, BytesIO]:
        """
        Gets an instance of the memory file class for the disk type.

        :param initial_value:   The initial contents for the file.
        :return:                A new memory file.
        """
        return cls._get_memory_file_class()(initial_value)

    @classmethod
    def _open(cls, filename: str, mode: str) -> IO[DiskType]:
        """
        Opens the given file is the given mode.

        :param filename:    The file to open.
        :param mode:        The base open mode (r/w).
        :return:            The file stream.
        """
        # Make sure a valid mode is given
        if mode not in ('r', 'w', 'a'):
            raise ValueError(f"Mode must be 'r', 'w' or 'a', not '{mode}'")

        # Get the open function to use
        open_func, mode = cls._get_open_func(filename, mode)

        # Open and return the file
        return open_func(filename, mode)

    @classmethod
    def _get_memory_file_class(cls) -> Union[Type[StringIO], Type[BytesIO]]:
        """
        Gets the IO class to use for the disk type that the reader class
        supports.

        :return:    Either StringIO or BytesIO.
        """
        return cls._map_disk_type(StringIO, BytesIO)

    @classmethod
    def _get_open_func(cls, filename: str, mode: str) -> Tuple[OpenFunc, str]:
        """
        Gets the open function to use to open the given filename.

        :param filename:    The file to open.
        :param mode:        The base open mode (r/w).
        :return:            The open function, and the read-mode flag to use.
        """
        if filename.endswith('.gz'):
            return gzip.open, cls._map_disk_type('{}t', '{}').format(mode)
        else:
            return open, cls._map_disk_type('{}', '{}b').format(mode)

    @classmethod
    def _map_disk_type(cls, str_, bytes_):
        """
        Returns str_ if DiskType is str and bytes_ if DiskType is bytes.
        Raises an error if it's neither.

        :param str_:    The return value when DiskType is str.
        :param bytes_:  The return type when DiskType is bytes.
        :return:        Either str_ or bytes_.
        """
        if cls._disk_type is str:
            return str_
        elif cls._disk_type is bytes:
            return bytes_
        else:
            raise TypeError(f"DiskType must be str or bytes, not {cls._disk_type}")

    @classmethod
    def _check_file(cls, file: IO[DiskType]):
        """
        Utility method that checks if the file is valid. Raises an error if it's not.

        :param file:  The file.
        """
        if file.closed:
            raise ValueError(f"File is closed")

    def __enter__(self):
        self.ensure_file()
        self._file.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ensure_file()
        return self._file.__exit__(exc_type, exc_val, exc_tb)



from typing import TypeVar

from ._get_argument_to_typevar import get_argument_to_typevar


class TypeVarProperty:
    """
    Property class which caches the dynamic type of a type-variable
    for a class.
    """
    def __init__(self, typevar: TypeVar):
        self._typevar: TypeVar = typevar
        self._base_class = None
        self._cache = {}

    def __get__(self, instance, owner):
        # Get the typevar argument if it's not cached
        if owner not in self._cache:
            self._cache[owner] = get_argument_to_typevar(owner, self._base_class, self._typevar)

        return self._cache[owner]

    def __set_name__(self, owner, name):
        # Prevents reassignment to parameterised generic base class
        if self._base_class is not None:
            return

        self._base_class = owner


import re
from typing import IO, Set, Iterable, Optional, Tuple

from ._error import ARFFError
from . import constants

# The set of characters that can delimit a data value
DATA_DELIMITERS: Set[str] = {'\t', ',', '\n'}

# The set of quotation symbols
QUOTES: Set[str] = {'\"', '\''}

# Regular expression for matching a name (for a relation or attribute)
# (possibly surrounded by quotes)
NAME_PATTERN: str = '''
    (?P<quote>['"])?            # An optional beginning quote
    (?![{},%])                  # Must not start with '{', '}', ',' or '%'
    (?(quote)                   # If we are in quotes...
        .+?                     # Match any characters up to the first end-quote
        |                       # If we are not in quotes...
        (\\S)+                  # Match any non-whitespace characters
    )
    (?(quote)(?P=quote))        # Match the same ending quote as was found at the beginning (if there was one)
    '''

# Regular expression for matching the name in a line
# (possibly surrounded by quotes)
NOMINAL_VALUE_PATTERN: str = '''
    (?P<quote>['"])?            # An optional beginning quote
    (?(quote)                   # If we are in quotes...
        .+?                     # Match any characters up to the close quotation
        |                       # If we are not in quotes...
        [^\\s,}]+               # Match to the next space or delimiter
    )
    (?(quote)(?P=quote))        # Match the same ending quote as was found at the beginning (if there was one)
    '''

# Matches the attribute type portion of an attribute declaration
ATTRIBUTE_TYPE_PATTERN: str = '''
    (numeric|integer|real|string|date|{.*?})    # A type keyword or a nominal specification
    '''

# Regular expression which matches a data row value and its delimiter
DATA_VALUE_PATTERN: str = '''
    (?P<quote>['"])?            # An optional beginning quote
    .*?                         # Any characters can exist in the value (will be checked separately)
    (?(quote)(?P=quote))        # Match the same ending quote as was found at the beginning (if there was one)
    \\s*                        # Some possible trailing whitespace
    (\\t|,|\\n)                 # The delimiter is a tab, a comma, or a new line
    '''

# Regular expression which matches a date-format string
# TODO: Implement properly. csterling
DATE_FORMAT_PATTERN: str = '''
    .*          # Just match anything for now
    '''


def consume(line: str,
            pattern: str,
            ignore_leading_whitespace: bool = True,
            case_insensitive: bool = True) -> Tuple[Optional[str], str]:
    """
    Consumes the beginning portion of the line that is matched by the given pattern.

    :param line:                        The line to consume from the start of.
    :param pattern:                     The pattern to match for consumption.
    :param ignore_leading_whitespace:   Whether to drop any leading whitespace.
    :param case_insensitive:            Whether to ignore case.
    :return:                            The consumed part of the line, and the remainder.
    """

    # Remove leading whitespace if ignoring
    if ignore_leading_whitespace:
        line = line.lstrip()

    # Initialise the flags
    flags = re.VERBOSE
    if case_insensitive:
        flags |= re.IGNORECASE

    # Match the pattern
    match = re.match(pattern, line, flags)

    # Return the matched part of the line and the remainder
    if match:
        return line[:match.end()], line[match.end():]
    else:
        return None, line


def read_till_found(file: IO[str],
                    words: Set[str],
                    search_comments: bool = False,
                    case_insensitive: bool = True) -> str:
    """
    Reads lines from the ARFF file until a line is found that contains at least
    one of the given search words.

    :param file:                The ARFF file to read from.
    :param words:               The set of words to search for.
    :param search_comments:     Whether to allow the search to look inside comments as well.
    :param case_insensitive:    Whether the search should be case-insensitive.
    :return:                    The line where the word was found.
    """

    # Lower-case all search words if doing case-insensitive search
    if case_insensitive:
        words = {word.lower() for word in words}

    # Keep reading lines until we run out or find the desired line
    line = None
    while line != "":
        # Read the next line from the file
        line = file.readline()

        # Skip blank lines
        if is_whitespace_only(line):
            continue

        # Skip comments if we are not searching them
        if not search_comments:
            if is_comment_line(line):
                continue

        # Lower-case the searched line if doing case-insensitive search
        search_line = line.lower() if case_insensitive else line

        # If it contains one of the words, return it
        for word in words:
            if search_line.find(word) != -1:
                return line

    # None of the words were found
    raise KeywordNotFoundError(words)


def remove_quotes(string: str) -> str:
    """
    Removes quotes from around a string, if they are present.

    :param string:  The string to remove quotes from.
    :return:        The string without quotes.
    """

    # If starts and ends with double-quotes, remove them
    if string.startswith('\"') and string.endswith('\"'):
        string = string[1:-1]

    # If starts and ends with single-quotes, remove them
    elif string.endswith('\'') and string.endswith('\''):
        string = string[1:-1]

    # Return the (unquoted) string
    return string


def is_comment_line(line: str) -> bool:
    """
    Checks whether the given line is a comment line.

    :param line:    The line to check.
    :return:        True if the line is a comment line, False if not.
    """

    return line_starts_with(line, constants.COMMENT_SYMBOL, False)


def remove_keyword(line: str, keyword: str) -> str:
    """
    Removes the given keyword from the beginning of the line.

    :param line:        The line to remove the keyword from.
    :param keyword:     The keyword to remove.
    :return:            The line with the keyword removed.
    """

    # The line should start with the given keyword
    assert line_starts_with(line, keyword), line + ' does not start with keyword: ' + keyword

    # Return the line with the keyword removed
    return line[len(keyword):]


def is_whitespace_only(line: str) -> bool:
    """
    Checks if the given line contains only whitespace.

    :param line:    The line to check.
    :return:        True if the line contains only whitespace, False if not.
    """

    return line.strip() == ''


def line_starts_with(line: str, string: str, case_insensitive: bool = True) -> bool:
    """
    Checks if the line starts with the given string.

    :param line:                The line to check.
    :param string:              The string to look for.
    :param case_insensitive:    Whether the check should disregard case.
    :return:                    True if the line starts with the given string, False otherwise.
    """

    # Remove case information if case-insensitive
    if case_insensitive:
        line = line.lower()
        string = string.lower()

    # Return the result of the check
    return line.startswith(string)


class KeywordNotFoundError(ARFFError):
    """
    Exception where an expected keyword was never found in the ARFF file.
    """
    def __init__(self, keywords: Iterable[str]):
        # Format the message
        message = "Couldn't find line containing any of: " + ", ".join(keywords)
        super().__init__(message)


class UnrecognisedContentError(ARFFError):
    """
    Exception when part of a line cannot be parsed because it is not valid ARFF content.
    """
    def __init__(self, content: str, line: str):
        super().__init__("Unrecognised content '" + content + "' in line: " + line)



from typing import Optional, Iterator

from .._typing import T


class ConstantIterator(Iterator[T]):
    """
    Class which implements the iterator interface, but always returns a single value.
    """
    def __init__(self, value: T, num_iterations: Optional[int] = None):
        self._value: T = value
        self._num_iterations: Optional[int] = num_iterations

    def __next__(self) -> T:
        if self._num_iterations == 0:
            raise StopIteration
        elif self._num_iterations is not None:
            self._num_iterations -= 1

        return self._value



from typing import IO

from .._Serialiser import Serialiser


class BoolSerialiser(Serialiser[bool]):
    """
    Serialiser which serialises a boolean value to a single byte,
    with 0x00 representing False and 0x01 representing True.
    """
    def __init__(self,
                 true_byte: bytes = b'\x01',
                 false_byte: bytes = b'\x00'):
        # Make sure the true and false bytes are 1-byte long
        if len(true_byte) != 1 or len(false_byte) != 1:
            raise ValueError(f"true_byte and false_byte must be single-byte values")

        # Make sure true and false aren't the same
        if true_byte == false_byte:
            raise ValueError(f"true_byte and false_byte are indistinguishable")

        self._true_byte = true_byte
        self._false_byte = false_byte

    def _check(self, obj: bool):
        # Make sure we got a tuple of bools
        if not isinstance(obj, bool):
            raise TypeError(f"{BoolSerialiser.__name__} serialises bools, got {type(obj)}")

    def _serialise(self, obj: bool, stream: IO[bytes]):
        stream.write(self._true_byte if obj else self._false_byte)

    def _deserialise(self, stream: IO[bytes]) -> bool:
        byte = stream.read(1)

        if byte == self._true_byte:
            return True
        elif byte == self._false_byte:
            return False
        else:
            raise ValueError(f"Received invalid byte '{byte}'")



import struct
from typing import IO

from .._Serialiser import Serialiser


class FloatSerialiser(Serialiser[float]):
    def __init__(self,
                 little_endian: bool = True,
                 double_precision: bool = True):
        self._format = f"{'<' if little_endian else '>'}{'d' if double_precision else 'f'}"
        self._num_bytes: int = 8 if double_precision else 4

    def _check(self, obj: float):
        # Make sure the object really is a float
        if not isinstance(obj, float):
            raise TypeError(f"Received non-float value of type {type(obj).__name__}")

    def _serialise(self, obj: float, stream: IO[bytes]):
        stream.write(struct.pack(self._format, obj))

    def _deserialise(self, stream: IO[bytes]) -> float:
        return struct.unpack(self._format, stream.read(self._num_bytes))[0]



from shlex import quote, split

from .._typing import OptionsList, is_options_list


def to_cli_string(options_list: OptionsList) -> str:
    """
    Forms an options list into a single CLI string.

    :param options_list:    The options list.
    :return:                The CLI string.
    """
    # Make sure it is an options list
    if not is_options_list(options_list):
        raise ValueError(f"Expected options list, got: {options_list}")

    options_list = [option if '"' in option else option.replace("'", '"')
                    for option in options_list]

    return " ".join(quote(option) for option in options_list)


def from_cli_string(cli_string: str) -> OptionsList:
    """
    Splits a CLI string into an options list.

    :param cli_string:  The CLI string.
    :return:            The options list.
    """
    return split(cli_string)



from enum import Enum, unique
from numbers import Real
from typing import Optional, Any, Union


@unique
class DataType(Enum):
    """
    The type of a field in a report.
    """
    STRING = "S"
    NUMERIC = "N"
    BOOLEAN = "B"
    UNKNOWN = "U"

    def to_display(self) -> str:
        """
        Returns the display string.

        :return:    The display string.
        """
        return self.value

    def to_raw(self) -> str:
        """
        Returns the raw enum string.

        :return:    The raw enum string.
        """
        return self.name

    def __str__(self) -> str:
        return self.to_display()

    @classmethod
    def value_of(cls, string: str) -> "DataType":
        """
        Gets the enum constant with the given name.

        :param string:  The name of the enum member.
        :return:        The enum instance.
        """
        return cls[string]

    @classmethod
    def parse(cls, string: str) -> Optional["DataType"]:
        """
        Parses the given string and returns the associated enum.

        :param string:  The string to parse.
        :return:        The enum, or None if not found.
        """
        # Attempt to parse enum name first
        try:
            return cls.value_of(string)
        except Exception:
            pass

        # Then try one-letter code
        try:
            return DataType(string)
        except Exception:
            pass

        return None

    @classmethod
    def guess_type(cls, obj: Any) -> "DataType":
        """
        Guesses the datatype from the object. In the case of string, checks
        if the string represents a number, then a boolean, then defaults to
        string.

        :param obj: The object to inspect.
        :return:    The datatype.
        """
        # If the object is a string, try to convert it
        if isinstance(obj, str):
            for datatype in (DataType.NUMERIC, DataType.BOOLEAN):
                try:
                    obj = datatype.convert(obj)
                except Exception:
                    pass

        # If the object is a number, return NUMERIC
        if isinstance(obj, Real):
            return DataType.NUMERIC

        # If the object is a bool, return BOOLEAN
        if isinstance(obj, bool):
            return DataType.BOOLEAN

        # Default to string
        return DataType.STRING

    def convert(self, obj: Any) -> Union[str, float, bool]:
        """
        Converts a value into the correct internal type for this
        datatype.

        :param obj:     The value to convert.
        :return:        The converted value.
        """
        if self is DataType.BOOLEAN:
            return bool(obj)
        elif self is DataType.NUMERIC:
            return float(obj)
        else:
            return str(obj)



"""
Module for count-translation functionality for the CountOption class.
"""
from typing import Dict

from ...meta.code_repr import CodeRepresentable, from_init, CodeRepresentation


class TranslationTable(CodeRepresentable):
    """
    Reversible translation between integer spaces.
    """
    def __init__(self, *translated_values: int):
        # Must be at least 2 values
        if len(translated_values) < 2:
            raise ValueError("Must provide at least 2 values")

        # The translation must be reversible
        if len(set(translated_values)) != len(translated_values):
            raise ValueError("Translated values are not unique")

        self._code_representation: CodeRepresentation = from_init(self, locals())

        self._forward_translation: Dict[int, int] = {index: value for index, value in enumerate(translated_values)}
        self._reverse_translation: Dict[int, int] = {value: key for key, value in self._forward_translation.items()}

    def translate_forward(self, value: int) -> int:
        # Cap the value to the range of the forward translation
        value = min(max(0, value), len(self._forward_translation) - 1)

        return self._forward_translation[value]

    def can_translate_reverse(self, value: int) -> bool:
        """
        Whether the given value is valid under reverse translation.

        :param value:   The value to check.
        :return:        True if the value can be reverse-translated.
        """
        return value in self._reverse_translation

    def translate_reverse(self, value: int) -> int:
        # Make sure the value is translatable
        if not self.can_translate_reverse(value):
            raise ValueError(f"Value not in translation table: {value}")

        return self._reverse_translation[value]

    def code_repr(self) -> CodeRepresentation:
        return self._code_representation



"""
Utility functions for working with iterables and iterators.
"""
from ._ConstantIterator import ConstantIterator
from ._functions import (
    safe_iter,
    flatten,
    invert_indices,
    extract_by_index,
    is_iterable,
    first,
    random,
    all_meet_predicate,
    any_meets_predicate,
    count
)
from ._metadata import with_metadata, zip_metadata, get_metadata, unzip_metadata



from typing import Type


def fully_qualified_name(cls: Type) -> str:
    """
    Gets the fully-qualified name of a class.

    :param cls:     The class.
    :return:        The fully-qualified name.
    """
    return f"{cls.__module__}.{cls.__qualname__}"



"""
Package for working with code-representations of objects.
"""
from ._CodeRepresentable import CodeRepresentable
from ._error import CodeRepresentationError, IsNotCodeRepresentableValue, IsNotCodeRepresentableType, ConflictingImports
from ._functional import code_repr, get_code_repr_function
from ._typing import ImportDict, CodeRepresentation, CodeReprFunction
from ._utilities import combine_import_dicts, from_init, get_code, get_import_dict



from ._BoolSerialiser import BoolSerialiser
from ._FixedLengthStringSerialiser import FixedLengthStringSerialiser
from ._FloatSerialiser import FloatSerialiser
from ._IntSerialiser import IntSerialiser
from ._LengthPrefixedStringSerialiser import LengthPrefixedStringSerialiser
from ._NullTerminatedStringSerialiser import NullTerminatedStringSerialiser
from ._SecondsSinceEpochSerialiser import SecondsSinceEpochSerialiser
from ._StringSerialiser import StringSerialiser, NULL_BYTE



"""
Module for code_repr functions for builtin types.
"""
from enum import Enum
from typing import Tuple, Iterator, Callable, Any

from .._import_code import get_import_code
from ._typing import CodeRepresentation


def type_code_repr(cls: type) -> CodeRepresentation:
    """
    The code_repr function for types.

    :param cls:     The type to get the code representation for.
    :return:        The code representation of the type.
    """
    # Get the import code for the class
    import_code = get_import_code(cls, alias_inner_class=False)

    return {cls.__qualname__: import_code}, cls.__qualname__


def primitive_code_repr(primitive) -> CodeRepresentation:
    """
    Gets the code representation for a primitive (float, str, etc.).

    :param primitive:   A primitive value.
    :return:            The value's code representation.
    """
    # No imports required, and code can be got from repr
    return {}, repr(primitive)


def container_code_repr(open_brace: str,
                        close_brace: str,
                        element_iterator: Iterator[Tuple],
                        formatter: Callable[[Any], str]) -> CodeRepresentation:
    """
    Base functionality for getting the code-representation
    of one of the container types.

    :param open_brace:                  The start of the code representation.
    :param close_brace:                 The end of the code representation.
    :param element_iterator:            An iterator over tuples containing the elements of the container.
    :param formatter:                   A formatter which formats the code of an element
                                        into its container representation.
    :return:                            The code representation of the container.
    :raises CodeRepresentationError:    If the container can't be represented.
    """
    # Local imports to avoid circularity errors
    from ._functional import code_repr
    from ._utilities import combine_import_dicts, get_import_dict, get_code

    # Initialise the code representation with an open bracket and empty imports
    import_dict = {}
    code = open_brace

    # Process each element in turn
    first = True
    for element in element_iterator:
        # Get the code-representation for the element
        element_reprs = tuple(map(code_repr, element))

        # Add the element imports to our own
        import_dict = combine_import_dicts(import_dict, *map(get_import_dict, element_reprs))

        # Insert a comma for every item after the first
        if first:
            first = False
        else:
            code += ", "

        # Add the key-value pair representation
        code += formatter(*map(get_code, element_reprs))

    # Close out the braces
    code += close_brace

    return import_dict, code


def dict_code_repr(value: dict) -> CodeRepresentation:
    """
    The code_repr function for dictionaries.

    :param value:   The dictionary.
    :return:        The code-representation of the dictionary.
    """
    return container_code_repr("{",
                               "}",
                               map(tuple, value.items()),
                               lambda key, val: f"{key}: {val}")


def list_code_repr(value: list) -> CodeRepresentation:
    """
    Gets the code representation for a list.

    :param value:   The list.
    :return:        It's code representation.
    """
    return container_code_repr("[",
                               "]",
                               ((el,) for el in value),
                               lambda el: el)


def tuple_code_repr(value: tuple) -> CodeRepresentation:
    """
    Gets the code representation for a tuple.

    :param value:   The tuple.
    :return:        It's code representation.
    """
    return container_code_repr("(",
                               ")",
                               ((el,) for el in value),
                               lambda el: el)


def set_code_repr(value: set) -> CodeRepresentation:
    """
    Gets the code representation for a set.

    :param value:   The set.
    :return:        It's code representation.
    """
    return container_code_repr("{",
                               "}",
                               ((el,) for el in value),
                               lambda el: el)


def frozenset_code_repr(value: frozenset) -> CodeRepresentation:
    """
    Gets the code representation for a frozenset.

    :param value:   The frozenset.
    :return:        It's code representation.
    """
    return container_code_repr("frozenset({",
                               "})",
                               ((el,) for el in value),
                               lambda el: el)


def enum_code_repr(value: Enum):
    """
    The code_repr function for enumerations.

    :param value:   The enum being represented.
    :return:        The code representation of the enum.
    """
    # Get the enum type
    enum_type = type(value)

    # Get the import code for the enumeration
    import_code = get_import_code(enum_type, alias_inner_class=False)

    return {enum_type.__qualname__: import_code}, f"{enum_type.__qualname__}.{value.name}"



"""
Functional interface to code representations.
"""
from enum import Enum
from typing import Optional, Any

from ._builtins import type_code_repr, enum_code_repr
from ._CodeRepresentable import CodeRepresentable
from ._error import IsNotCodeRepresentableType, CodeRepresentationError, IsNotCodeRepresentableValue
from ._honorary_members import get_honorary_code_repr_function
from ._typing import CodeReprFunction, CodeRepresentation


def get_code_repr_function(cls: type) -> Optional[CodeReprFunction]:
    """
    Gets the code_repr function for the given type.

    :param cls:     The type to get the function for.
    :return:        The code_repr function, or None if none available.
    """
    # Check for honorary members
    honorary_code_repr_function = get_honorary_code_repr_function(cls)
    if honorary_code_repr_function is not None:
        return honorary_code_repr_function

    # Check for CodeRepresentable sub-types
    if issubclass(cls, CodeRepresentable):
        return cls.code_repr

    # Types are also code-representable
    if issubclass(cls, type):
        return type_code_repr

    # Enumerations are also code representable
    if issubclass(cls, Enum):
        return enum_code_repr

    return None


def code_repr(obj: Any) -> CodeRepresentation:
    """
    Gets the code representation of the object.

    :return:                                The code representation.
    :raises IsNotCodeRepresentableType:     If the value's type is not a code-representable type.
    :raises IsNotCodeRepresentableValue:    If the value is not code-representable.
    """
    # Get the object's type
    cls = type(obj)

    # Get the code_repr function
    code_repr_function = get_code_repr_function(cls)

    # If it doesn't have a function, it's not code-representable
    if code_repr_function is None:
        raise IsNotCodeRepresentableType(cls)

    # Try to get the code-representation of the object
    try:
        return code_repr_function(obj)

    # If it fails, the value is not code-representable
    except CodeRepresentationError as e:
        raise IsNotCodeRepresentableValue(obj) from e



def sequence(*decorators):
    """
    Helper method which creates a decorator that applies the given
    sub-decorators. Decorators are applied in reverse order given.

    @decorator_sequence(dec_1, dec_2, ...)
    def function(...):
        ...

    is equivalent to:

    @dec_1
    @dec_2
    ...
    def function(...):
        ...

    :param decorators:  The sub-decorators.
    :return:            A function which applies the given decorators.
    """
    def apply(obj):
        for dec in reversed(decorators):
            obj = dec(obj)
        return obj

    return apply



"""
Package for working with command-line options.
"""
from ._ArgumentParserConfigurer import ArgumentParserConfigurer
from ._CLIFactory import CLIFactory
from ._CLIInstantiable import CLIInstantiable
from ._CLIRepresentable import CLIRepresentable, cli_repr, from_cli_repr
from ._OptionHandler import OptionHandler
from ._OptionValueHandler import OptionValueHandler
from ._typing import OptionsList, is_options_list



from ._CSVFile import CSVFile


def save(csv: CSVFile, filename: str, encoding: str = 'utf-8'):
    """
    Saves the given CSV file to disk.

    :param csv:        The CSV file to save.
    :param filename:    The filename to save the file under.
    :param encoding:    The string encoding to use.
    """
    with open(filename, 'w', encoding=encoding) as file:
        file.write(str(csv))



"""
Module defining the honorary members of the CodeRepresentable class.
"""
from typing import Optional, Dict

from ._builtins import (
    primitive_code_repr,
    dict_code_repr,
    list_code_repr,
    tuple_code_repr,
    set_code_repr,
    frozenset_code_repr
)
from ._typing import CodeReprFunction

# The honorary code-representable types, and their code_repr methods.
HONORARY_MEMBERS: Dict[type, CodeReprFunction] = {
    str: primitive_code_repr,
    bytes: primitive_code_repr,
    bytearray: primitive_code_repr,
    complex: primitive_code_repr,
    float: primitive_code_repr,
    int: primitive_code_repr,
    bool: primitive_code_repr,
    None: primitive_code_repr,
    range: primitive_code_repr,
    slice: primitive_code_repr,
    dict: dict_code_repr,
    list: list_code_repr,
    tuple: tuple_code_repr,
    set: set_code_repr,
    frozenset: frozenset_code_repr
}


def get_honorary_code_repr_function(cls: type) -> Optional[CodeReprFunction]:
    """
    Gets the code_repr function for an honorary type.

    :param cls:     The type.
    :return:        The code_repr function if it exists, otherwise None.
    """
    return HONORARY_MEMBERS.get(cls, None)



"""
Constants related to ADAMS reports.
"""
# The file extension of a report file
EXTENSION: str = ".report"

# The suffix for the names of properties which represent the datatype of another property
DATATYPE_SUFFIX: str = "\tDataType"

# The parent ID property
PROPERTY_PARENTID: str = "Parent ID"

# The database ID used when none is given
NO_ID: int = -1

# Field: Dummy report (in case there was no quantitation report and a
# dummy report was generated automatically)
FIELD_DUMMYREPORT: str = "Dummy report"

# Field: Excluded (= dodgy)
FIELD_EXCLUDED: str = "Excluded"



from inspect import Signature, signature, Parameter
from typing import Any, Dict


def non_default_kwargs(function, locals_) -> Dict[str, Any]:
    """
    Gets a dictionary of all keyword-only arguments which
    have non-default values.

    E.g.
    def my_func(*, kw1=None, kw2=None, kw3=None):
        return non_default_kwargs(my_func, locals())

    nd_kwargs = my_func(kw1=1, kw3=3)

    assert nd_kwargs == {"kw1": 1, "kw3": 3}

    :param function:    The function being called.
    :param locals_:     The locals in the function.
    :return:            The mapping from keyword to value for
                        all non-default keyword arguments.
    """
    # Get the function signature
    function_signature: Signature = signature(function)

    return {name: locals_[name]
            for name, parameter in function_signature.parameters.items()
            if parameter.kind == Parameter.KEYWORD_ONLY
            and parameter.default is not Parameter.empty
            and locals_[name] is not parameter.default}



from typing import TypeVar, Dict, Callable, Tuple, Any, NoReturn, Optional, Type

from ..meta import fully_qualified_name

# Generic type variable
T = TypeVar("T")

# The types of the 'cli_repr'/'from_cli_repr' methods
CLI_REPR_METHOD_TYPE = Callable[[T], str]
FROM_CLI_REPR_METHOD_TYPE = Callable[[str], T]


# For use when honarary member methods aren't implemented (yet)
def NOT_IMPLEMENTED(method_name: str) -> Callable[[Any], NoReturn]:
    def method(*args, **kwargs):
        raise NotImplementedError(method_name)

    return method


def type_from_cli_repr(cli_string: str) -> type:
    """
    Imports a class from its CLI representation.
    TODO: Move to meta package

    :param cli_string:  The CLI representation of the class.
    :return:            The class.
    """
    locals_ = dict(locals())
    parts = cli_string.split(".")
    for i in range(len(parts) - 1, 0, -1):
        code = f"from {'.'.join(parts[:i])} import {parts[i]} as result"
        if i != len(parts) - 1:
            code += f"\nresult = result.{'.'.join(parts[i+1:])}"
        try:
            exec(code, globals(), locals_)
        except ImportError:
            pass
        else:
            return locals_["result"]

    raise ImportError(f"Couldn't import class '{cli_string}'")


# Honorary CLI-representable types and their 'cli_repr'/'from_cli_repr' methods
HONORARY_MEMBERS: Dict[Type[T], Tuple[CLI_REPR_METHOD_TYPE[T], FROM_CLI_REPR_METHOD_TYPE[T]]] = {
    int: (repr, int),
    str: (repr, lambda cli_string: cli_string),
    float: (repr, float)
}


class CLIRepresentable:
    """
    Interface for objects which can represent themselves
    as a value on the command-line.
    """
    def cli_repr(self) -> str:
        """
        Gets the CLI representation of this object.

        :return:        The CLI string.
        """
        raise NotImplementedError(CLIRepresentable.cli_repr.__qualname__)

    @classmethod
    def from_cli_repr(cls: Type[T], cli_string: str) -> T:
        """
        Gets an instance of the type from its CLI representation.

        :param cli_string:  The CLI representation.
        :return:            The instance.
        """
        raise NotImplementedError(CLIRepresentable.from_cli_repr.__qualname__)


def get_cli_repr_functions(cls: Type[T]) -> Optional[Tuple[CLI_REPR_METHOD_TYPE[T], FROM_CLI_REPR_METHOD_TYPE[T]]]:
    """
    Gets the cli_repr/from_cli_repr functions for a given type.

    :param cls:     The type to get the functions for.
    :return:        The pair of functions, or None if none available.
    """
    # See if the type is an honorary member
    if cls in HONORARY_MEMBERS:
        return HONORARY_MEMBERS[cls]

    # If the class inherits from CLIRepresentable, return its cli_repr method
    if issubclass(cls, CLIRepresentable):
        return cls.cli_repr, cls.from_cli_repr

    # If the class is a meta-class (i.e. it's values are classes),
    # then represent it by it's fully-qualified name, and parse
    # it by importing it
    if issubclass(cls, type):
        return fully_qualified_name, type_from_cli_repr

    return None


def get_cli_repr_function(value: T) -> Optional[CLI_REPR_METHOD_TYPE[T]]:
    """
    Gets the function to call to get the CLI representation of a value.

    :param value:   The value to get the cli_repr function for.
    :return:        The cli_repr function, or None if none available.
    """
    # Get the value's type
    cls = type(value)

    # Get the functions for the type
    cli_repr_functions = get_cli_repr_functions(cls)

    # If the functions exist, return the cli_repr function
    if cli_repr_functions is not None:
        return cli_repr_functions[0]

    return None


def get_from_cli_repr_function(cls: Type[T]) -> Optional[CLI_REPR_METHOD_TYPE[T]]:
    """
    Gets the function to call to parse the CLI representation of a value.

    :param cls:     The type to get the from_cli_repr function for.
    :return:        The from_cli_repr function, or None if none available.
    """
    # Get the functions for the type
    cli_repr_functions = get_cli_repr_functions(cls)

    # If the functions exist, return the from_cli_repr function
    if cli_repr_functions is not None:
        return cli_repr_functions[1]

    return None


def is_cli_representable_type(cls: Type[T]) -> bool:
    """
    Checks if the given class is a CLI-representable type.

    :param cls:     The type to check.
    :return:        True if the type is CLI-representable.
    """
    return get_cli_repr_functions(cls) is not None


def is_cli_representable(value: T) -> bool:
    """
    Checks if the given value is CLI-representable.

    :param value:   The value to check.
    :return:        True if the value is CLI-representable.
    """
    return is_cli_representable_type(type(value))


def cli_repr(value: T) -> str:
    """
    Gets the CLI representation of a CLI-representable value.

    :param value:   The value.
    :return:        The CLI string.
    :raises TypeError:  If the type is not CLI-representable.
    """
    # Get the cli_repr function for the value
    cli_repr_function = get_cli_repr_function(value)

    # If it doesn't have a function, it's not CLI-representable
    if cli_repr_function is None:
        raise TypeError(f"Values of type '{type.__name__}' are not CLI-representable")

    return cli_repr_function(value)


def from_cli_repr(cls: Type[T], cli_string: str) -> T:
    """
    Gets a value of a CLI-representable type from its CLI representation.

    :param cls:         The type of value to get.
    :param cli_string:  The CLI representation.
    :return:            The value.
    :raises TypeError:  If the type is not CLI-representable.
    """
    # Get the from_cli_repr function for the type
    from_cli_repr_function = get_from_cli_repr_function(cls)

    # If it doesn't have a function, it's not CLI-representable
    if from_cli_repr_function is None:
        raise TypeError(f"'{cls.__qualname__}' is not a CLI-representable type")

    return from_cli_repr_function(cli_string)



import logging


def create_standard_library_root_logger(library_root_package: str) -> logging.Logger:
    """
    Creates a root logger for libraries. This is a logger
    with a single null handler attached to it so that libraries
    can be used without logging set up.

    :param library_root_package:    The package name of the library.
    :return:                        The created logger.
    """
    # Get the logger for the package
    library_root_logger: logging.Logger = logging.getLogger(library_root_package)

    # Check that the logger isn't already configured
    if len(library_root_logger.handlers) > 0:
        library_root_logger.warning(f"Attempted to reconfigure library root logger '{library_root_package}'")
    else:
        library_root_logger.addHandler(logging.NullHandler())

    return library_root_logger



from typing import Dict, Any, Optional, List

from ....geometry import Rectangle, Polygon, Point
from ....logging import LoggingMixin
from . import constants


class LocatedObject(LoggingMixin):
    """
    Container for meta-data about located objects.
    """
    def __init__(self, x: int, y: int, width: int, height: int, **metadata: Any):
        # Fix negative width
        if width < 0:
            x += width
            width = -width

        # Fix negative height
        if height < 0:
            y += height
            height = -height

        self.x: int = x
        self.y: int = y
        self.width: int = width
        self.height: int = height
        self.metadata: Dict[str, Any] = metadata

        self._actual_rectangle: Optional[Rectangle] = None
        self._actual_polygon: Optional[Polygon] = None

    def get_clone(self) -> "LocatedObject":
        """
        Returns a clone of the object.

        :return:    The clone.
        """
        return LocatedObject(self.x, self.y, self.width, self.height, **self.metadata.copy())

    def get_index_string(self) -> Optional[str]:
        """
        Returns the index string of the object.

        :return:    The index, None if not available.
        """
        return str(self.metadata[constants.KEY_INDEX]) if constants.KEY_INDEX in self.metadata else None

    def get_index(self) -> int:
        """
        Returns the index of an object.

        :return:    The index, or -1 if not available.
        """
        # Get the string index
        index_string: str = self.get_index_string()

        # Try to parse the string
        if index_string is not None:
            try:
                return int(index_string)
            except Exception:
                self.logger.exception(f"Error parsing index string '{index_string}'")

        return -1

    def get_actual_rectangle(self) -> Optional[Rectangle]:
        """
        Returns the actual size rectangle.

        :return:     The actual size rectangle.
        """
        if self._actual_rectangle is None:
            self._actual_rectangle = self.get_rectangle()

        return self._actual_rectangle

    def get_actual_polygon(self) -> Optional[Polygon]:
        """
        Returns the actual size polygon.

        :return:    The actual size polygon.
        """
        if self._actual_polygon is None:
            self._actual_polygon = self.get_polygon()

        return self._actual_polygon

    def scale(self, scale: float):
        """
        Scales the actual size rectangle/polygon by the given factor.

        :param scale:   The scale factor.
        """
        self._actual_polygon = self.get_polygon(scale)
        self._actual_rectangle = self.get_rectangle(scale)

    def get_location(self) -> str:
        """
        Returns the quadrilateral location as a string.

        :return:    The string representation.
        """
        left: int = self.x
        top: int = self.y
        right: int = self.x + self.width - 1
        bottom: int = self.y + self.height - 1

        return f"{left} {top} {right} {top} {right} {bottom} {left} {bottom}"

    def has_polygon(self) -> bool:
        """
        Checks whether polygon meta-data is present.

        :return:    True if present.
        """
        return constants.KEY_POLY_X in self.metadata and constants.KEY_POLY_Y in self.metadata

    def get_rectangle(self, scale: float = 1.0) -> Rectangle:
        """
        Returns the object as a rectangle.

        :return:    The rectangle.
        """
        return Rectangle.x_y_w_h(int(scale * self.x),
                                 int(scale * self.y),
                                 int(scale * self.width),
                                 int(scale * self.height))

    def get_poly_coords(self, key: str) -> List[int]:
        """
        Returns the specified polygon coordinates.

        :param key:     constants.KEY_POLY_X or constants.KEY_POLY_Y
        :return:        The co-ordinates, 0-length if none available or failed to parse
        """
        # Check there is a polygon
        if not self.has_polygon():
            return []

        # Try to parse the coordinates
        try:
            return list(map(round, map(float, str(self.metadata[key]).split(","))))
        except Exception as e:
            self.logger.exception(f"Error parsing polygon coordinates for {key}")
            return []

    def get_polygon_x(self) -> List[int]:
        """
        Returns the x-coordinates of the polygon (if any).

        :return:    The x-coordinates.
        """
        return self.get_poly_coords(constants.KEY_POLY_X)

    def get_polygon_y(self) -> List[int]:
        """
        Returns the y-coordinates of the polygon (if any).

        :return:    The y-coordinates.
        """
        return self.get_poly_coords(constants.KEY_POLY_Y)

    def get_polygon(self, scale: float = 1.0) -> Optional[Polygon]:
        """
        Returns the polygon, if possible.

        :return:    A list of (x, y) pairs.
        """
        # Get the x and y coordinates
        xs: List[int] = self.get_polygon_x()
        ys: List[int] = self.get_polygon_y()

        # Make sure they are present and valid
        if len(xs) == 0 or len(xs) != len(ys):
            return None

        if scale != 1.0:
            for i in range(len(xs)):
                xs[i] = int(scale * xs[i])
                ys[i] = int(scale * ys[i])

        return Polygon(*(Point(x, y) for x, y in zip(xs, ys)))

    def set_polygon(self, polygon: Polygon):
        """
        Stores the polygon in the meta-data.

        :param polygon: The polygon.
        """
        self.metadata[constants.KEY_POLY_X] = ",".join(str(point.x) for point in polygon)
        self.metadata[constants.KEY_POLY_Y] = ",".join(str(point.y) for point in polygon)

    def make_fit(self, width: int, height: int) -> bool:
        """
        Ensures that the object fits within this region.

        :param width:   The width of the region.
        :param height:  The height of the region.
        :return:        True if the object was adjusted.
        """
        result: bool = False

        if self.x < 0:
            self.width += self.x
            self.x = 0
            result = True

        if self.x + self.width > width:
            self.width -= self.x + self.width - width
            result = True

        if self.y < 0:
            self.height += self.y
            self.y = 0
            result = True

        if self.y + self.height > height:
            self.height -= self.y + self.height - height
            result = True

        if self.has_polygon():
            px: List[int] = self.get_polygon_x()
            py: List[int] = self.get_polygon_y()
            padjusted: bool = False

            for i in range(len(px)):
                if px[i] < 0:
                    px[i] = 0
                    padjusted = True

                if px[i] >= width:
                    px[i] = width - 1
                    padjusted = True

                if py[i] < 0:
                    py[i] = 0
                    padjusted = True

                if py[i] >= height:
                    py[i] = height - 1
                    padjusted = True

            if padjusted:
                self.set_polygon(Polygon(*(Point(x, y) for x, y in zip(px, py))))
                result = True

        return result

    def overlap(self, other: "LocatedObject") -> bool:
        """
        Returns whether this and the other object overlap.

        :param other:   The other located object.
        :return:        True if they overlap.
        """
        return self.get_rectangle().intersects(other.get_rectangle())

    def overlap_rectangle(self, other: "LocatedObject") -> Rectangle:
        """
        Returns the overlapping rectangle.

        :param other:   The other located object.
        :return:        Rectangle if they overlap, otherwise None.
        """
        return self.get_rectangle().intersection(other.get_rectangle())

    def overlap_ratio(self, other: "LocatedObject") -> float:
        """
        Returns the overlap ratio (1 = full overlap, 0 = no overlap)

        :param other:   The other located object.
        :return:        The overlap ratio.
        """
        overlap: Rectangle = self.overlap_rectangle(other)

        if overlap is None:
            return 0.0

        return overlap.area() / self.get_rectangle().area()

    def compare_to(self, other: "LocatedObject") -> int:
        """
        Compares this object with the provided one. Bounding box,
        then polygon (if present).

        :param other:   The other object.
        :return:        If x/y/width/height are less than, equal to, or larger than
                        the other one.
        """
        result: int = self.x - other.x
        if result == 0:
            result = self.y - other.y
        if result == 0:
            result = self.width - other.width
        if result == 0:
            result = self.height - other.height
        if result == 0:
            result = self._compare_bool(self.has_polygon(), other.has_polygon())
        if result == 0:
            result = self._compare_list(self.get_polygon_x(), other.get_polygon_x())
        if result == 0:
            result = self._compare_list(self.get_polygon_y(), other.get_polygon_y())

        return result

    @staticmethod
    def _compare_bool(b1: bool, b2: bool):
        """
        Compares two booleans in the same way as Java's Boolean.compare method.

        :param b1:  The first bool.
        :param b2:  The second bool.
        :return:    -1, 0, +1 if b1 is smaller than, equal to, or larger than b2.
        """
        return 0 if b1 == b2 else 1 if b1 else -1

    @staticmethod
    def _compare_list(l1: List[int], l2: List[int]) -> int:
        """
        Compares two lists of integers.

        :param l1:  The first list.
        :param l2:  The second list.
        :return:    -1, 0, +1 if l1 is smaller than, equal to, or larger than l2.
        """
        result: int = len(l1) - len(l2)

        if result == 0:
            for i in range(len(l1)):
                result = l1[i] - l2[i]
                if result != 0:
                    break

        return result

    def equals(self, other: Any) -> bool:
        """
        Tests if this object is the same as the other one.

        :param other:   The other object.
        :return:        True if they are the same.
        """
        return isinstance(other, LocatedObject) and self.compare_to(other) == 0

    def hash_code(self) -> int:
        """
        Returns the hash-code of the rectangle.

        :return:    The hash.
        """
        return hash(self.get_rectangle())

    def to_string(self) -> str:
        """
        Returns a short description of the container.

        :return:    The description.
        """
        return f"x={self.x}, y={self.y}, w={self.width}, h={self.height}"

    def __hash__(self) -> int:
        return self.hash_code()

    def __str__(self) -> str:
        return self.to_string()



from ._typing import CodeRepresentation


class CodeRepresentable:
    """
    Base class for types that can represent themselves as a Python expression which,
    when evaluated, evaluates to a copy of themselves.
    """
    def code_repr(self) -> CodeRepresentation:
        """
        Gets the code representation of the object.

        :return:                            The code representation.
        :raise CodeRepresentationError:     If the object is not code-representable.
        """
        raise NotImplementedError(CodeRepresentable.code_repr.__qualname__)



from typing import Any, Optional, Type

from ...meta import non_default_kwargs
from ...meta.code_repr import CodeRepresentation, from_init
from ..._ClassRegistry import ClassRegistry
from .._typing import OptionsList
from ._Option import Option


class ClassOption(Option):
    """
    Option which selects a class from a class registry.
    Selection can only be made from aliases.
    """
    def __init__(self,
                 *flags: str,
                 registry: ClassRegistry,
                 default: Type = ...,
                 required: bool = ...,
                 metavar: str = ...,
                 help: str = ...):
        # Capture the code-representation of the option
        code_representation: CodeRepresentation = from_init(self, locals())

        # Save the registry
        self._registry: ClassRegistry = registry

        # Save our optionality
        self._optional: bool = not required if required is not ... else True

        # Make sure there are some choices to choose from
        minimum_required_choices = 1 if self._optional else 2
        choices = tuple(registry.aliases())
        if len(choices) < minimum_required_choices:
            raise ValueError("Not enough choices")

        # Extract the set of keyword arguments
        kwargs = non_default_kwargs(ClassOption.__init__, locals())

        super().__init__(code_representation,
                         *flags,
                         choices=choices,
                         **non_default_kwargs(ClassOption.__init__, locals()))

    def _namespace_value_to_internal_value(self, namespace_value: Optional[str]) -> Optional[type]:
        return self._registry.find(namespace_value) if namespace_value is not None else None

    def _internal_value_to_namespace_value(self, internal_value: Optional[type]) -> Optional[str]:
        return self._registry.get_alias(internal_value) if internal_value is not None else None

    def _namespace_value_to_options_list(self, namespace_value: Optional[str]) -> OptionsList:
        return [self._flags[0], namespace_value] if namespace_value is not None else []

    def _validate_internal_value(self, internal_value: Any):
        # If we're optional, can have the value None
        if self._optional and internal_value is None:
            return

        # Internal values are classes
        if not isinstance(internal_value, type):
            raise TypeError(f"Class-registry option expects classes for internal values")

        # Make sure it's an aliased class
        if self._registry.get_alias(internal_value) is None:
            raise ValueError(f"Type '{internal_value.__qualname__}' is not a valid type for this option")

    def _validate_namespace_value(self, namespace_value: Any):
        # If we're optional, can have the value None
        if self._optional and namespace_value is None:
            return

        # Namespace value should be a string otherwise
        if not isinstance(namespace_value, str):
            raise TypeError(f"Class-registry options expect string namespace values")

        # The string should be a registered alias
        if not self._registry.has_alias(namespace_value):
            raise NameError(f"Registry has no alias '{namespace_value}'")



from setuptools import setup, find_namespace_packages


def _read(f) -> bytes:
    """
    Reads in the content of the file.
    :param f: the file to read
    :type f: str
    :return: the content
    :rtype: str
    """
    return open(f, 'rb').read()


setup(
    name="wai.common",
    description="Python library with common functionality for other Waikato projects.",
    long_description=(
        _read('DESCRIPTION.rst') + b'\n' +
        _read('CHANGES.rst')).decode('utf-8'),
    url="https://github.com/waikato-datamining/wai-common",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Programming Language :: Python :: 3',
    ],
    license='MIT',
    package_dir={
        '': 'src'
    },
    packages=find_namespace_packages(where="src"),
    namespace_packages=[
        "wai"
    ],
    version="0.0.29",
    author='Corey Sterling',
    author_email='coreytsterling@gmail.com',
    install_requires=[
        "typing-inspect",
        "javaproperties"
    ],
    include_package_data=True
)



from abc import abstractmethod, ABC


class PrefixOnlyField(ABC):
    """
    Interface for fields that only have a prefix.
    """
    @abstractmethod
    def is_compound(self) -> bool:
        pass

    @abstractmethod
    def get_prefix(self) -> str:
        pass



"""
Package for utility functions for working with command-line options.
"""
from ._cli_string import to_cli_string, from_cli_string
from ._flags import (
    is_flag,
    is_flag_reason,
    is_short_flag,
    is_short_flag_reason,
    is_long_flag,
    is_long_flag_reason,
    flag_from_name
)
from ._translation import TranslationTable



from typing import Tuple

import math


class Point:
    """
    Represents an (x, y) coordinate in 2D integer Cartesian space.
    """
    def __init__(self, x: int, y: int):
        self.x: int = x
        self.y: int = y

    def __str__(self):
        return f"({self.x}, {self.y})"

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    @classmethod
    def colinear(cls, *points: "Point") -> bool:
        """
        Checks if the given points are co-linear.

        :param points:  The points to check.
        :return:        True if all points are co-linear.
        """
        # Any less than 3 points are always co-linear
        if len(points) < 3:
            return True

        # Calculate the slope between the first 2 points
        reference_slope = cls.slope(points[0], points[1])

        # All other slopes must be the same
        return all(cls.slope(points[0], point) == reference_slope for point in points[2:])

    @classmethod
    def slope(cls, a: "Point", b: "Point") -> Tuple[int, int]:
        """
        Calculates the slope between points a and b. If the slope
        is rise/run, then this function returns (rise, run), where
        run is always non-negative and gcd(rise, run) == 1.

        :param a:   One point.
        :param b:   The other point.
        :return:    The slope as (rise, run).
        """
        # If the two points are the same, slope is 0, 0
        if a == b:
            return 0, 0

        # Calculate the raw rise/run
        rise, run = b.y - a.y, b.x - a.x

        # Horizontal/vertical case
        if run == 0:
            return 1, 0
        if rise == 0:
            return 0, 1

        # Make sure run is non-negative
        if run < 0:
            rise = -rise
            run = -run

        # Calculate the GCD of rise and run
        slope_gcd = math.gcd(abs(rise), run)

        # Remove any common factors and return
        return rise // slope_gcd, run // slope_gcd



from typing import Type, Union


def get_import_code(cls: Type, indent: Union[int, str] = "", alias_inner_class: bool = True) -> str:
    """
    Gets the code to use to import a class.

    :param cls:                 The class.
    :param indent:              The indentation of the code.
    :param alias_inner_class:   Whether to add code to alias inner classes as outer classes.
    :return:                    The code.
    """
    # Can't get import code for closure classes
    if "<locals>" in cls.__qualname__:
        raise ValueError(f"Can't get import code for closure class '{cls.__qualname__}'")

    # Get the outer-most class to import
    outer_class = cls.__qualname__[:cls.__qualname__.index(".")] if "." in cls.__qualname__ else cls.__qualname__

    # Format the indentation string
    if isinstance(indent, int):
        indent = " " * indent

    # Get the code for importing the outer-most class
    code = f"{indent}from {cls.__module__} import {outer_class}"

    # If it's an inner class, alias it
    if cls.__name__ != outer_class and alias_inner_class:
        code += f"\n{indent}{cls.__name__} = {cls.__qualname__}"

    return code



# The keywords for the different sections of the file
RELATION_SECTION_KEYWORD = '@relation'
ATTRIBUTE_SECTION_KEYWORD = '@attribute'
DATA_SECTION_KEYWORD = '@data'

# Keywords for the different attribute datatypes
NUMERIC_ATTRIBUTE_KEYWORD = 'numeric'
INTEGER_ATTRIBUTE_KEYWORD = 'integer'
REAL_ATTRIBUTE_KEYWORD = 'real'
NOMINAL_ATTRIBUTE_KEYWORD = 'nominal'
DATE_ATTRIBUTE_KEYWORD = 'date'
STRING_ATTRIBUTE_KEYWORD = 'string'
RELATIONAL_ATTRIBUTE_KEYWORD = 'relational'

# The symbol that represents a missing value
MISSING_VALUE_SYMBOL = '?'

# The symbol that represents the beginning of a comment
COMMENT_SYMBOL = '%'



from typing import Any

from ._AbstractField import AbstractField
from ._Field import Field
from ._SuffixOnlyField import SuffixOnlyField
from ._DataType import DataType


class SuffixField(Field, SuffixOnlyField):
    """
    A compound field that only displays the second half of the name.
    """
    # The dummy suffix
    DUMMY_PREFIX: str = "DUMMY"

    def __init__(self, *args):
        if len(args) == 0:
            args = (None, DataType.UNKNOWN)
        elif len(args) > 2:
            raise ValueError("PrefixField only takes 0, 1 or 2 arguments to __init__")

        super().__init__(*args)

        if self.name is not None:
            if not self.is_compound():
                self.name = self.DUMMY_PREFIX + self.SEPARATOR + self.name
            else:
                self.name = self.DUMMY_PREFIX + self.SEPARATOR + self.get_suffix()

    def compare_to(self, other: Any) -> int:
        if other is None:
            return 1

        if not isinstance(other, AbstractField):
            return -1

        if self.get_suffix() is None and other.get_suffix() is None:
            return 0

        if self.get_suffix() is None:
            return -1

        if other.get_suffix() is None:
            return 1

        return (-1 if self.get_suffix().lower() < other.get_suffix().lower() else
                0 if self.get_suffix().lower() == other.get_suffix().lower() else
                1)

    def to_string(self) -> str:
        if self.name is None:
            return ""

        return self.get_suffix()

    def to_display_string(self) -> str:
        return self.to_string()

    @staticmethod
    def parse_field(string: str) -> Field:
        return SuffixField(Field.parse_field(string))



import datetime
from typing import IO

from .._Serialiser import Serialiser
from ._IntSerialiser import IntSerialiser


class SecondsSinceEpochSerialiser(Serialiser[datetime.datetime]):
    """
    Serialises a date/time as the number of seconds since a given epoch
    (as an integer).
    """
    def __init__(self,
                 epoch: datetime.datetime = datetime.datetime.fromtimestamp(0),
                 int_serialiser: Serialiser[int] = IntSerialiser()):
        self._epoch: datetime.datetime = epoch
        self._int_serialiser: Serialiser[int] = int_serialiser

    def _check(self, obj: datetime.datetime):
        # Make sure it really is a datetime
        if not isinstance(obj, datetime.datetime):
            raise TypeError(f"{SecondsSinceEpochSerialiser.__name__} serialises "
                            f"{datetime.datetime.__name__}s")

    def _serialise(self, obj: datetime.datetime, stream: IO[bytes]):
        # Get the number of seconds since the epoch
        num_seconds: int = int((obj - self._epoch).total_seconds())

        # Serialise it as an int
        self._int_serialiser.serialise(num_seconds, stream)

    def _deserialise(self, stream: IO[bytes]) -> datetime.datetime:
        # Deserialise the number of seconds since the epoch
        num_seconds: int = self._int_serialiser.deserialise(stream)

        # Convert to a time-delta
        time_delta = datetime.timedelta(seconds=float(num_seconds))

        return self._epoch + time_delta



from typing import Optional, Any

from ...meta.code_repr import CodeRepresentation, from_init
from ..constants import SHORT_FLAG_PREFIX
from ..util import is_short_flag, TranslationTable
from .._typing import OptionsList
from ._Option import Option


class CountOption(Option):
    """
    Option which counts the number of times its flag appears.

    Can optionally supply a translation table to translate from
    actual count values to some other set of integers (e.g. from
    counts to verbosity values).
    """
    def __init__(self,
                 *flags: str,
                 translation: Optional[TranslationTable] = None,
                 help: str = ...):
        # Capture the code-representation of the option
        code_representation: CodeRepresentation = from_init(self, locals())

        # Create the translation pair
        self._translation: Optional[TranslationTable] = translation

        super().__init__(code_representation,
                         *flags,
                         action="count",
                         default=translation.translate_forward(0) if translation is not None else 0,
                         help=help)

    def _namespace_value_to_internal_value(self, namespace_value: Any) -> int:
        return (
            self._translation.translate_forward(namespace_value)
            if self._translation is not None
            else namespace_value
        )

    def _internal_value_to_namespace_value(self, value: Any) -> int:
        return (
            self._translation.translate_reverse(value)
            if self._translation is not None
            else value
        )

    def _namespace_value_to_options_list(self, namespace_value: Any) -> OptionsList:
        # If count is 0, return no options
        if namespace_value == 0:
            return []

        # Try to use a short flag if available, otherwise just the first flag
        selected_flag = self.flags[0]
        short_flag = is_short_flag(selected_flag)
        if len(self.flags) == 2 and not short_flag:
            selected_flag = self.flags[1]
            short_flag = True

        # Iterate the flag count times
        if short_flag:
            return [SHORT_FLAG_PREFIX + selected_flag[1] * namespace_value]
        else:
            return [selected_flag] * namespace_value

    def _validate_internal_value(self, internal_value: Any):
        # Value must be an integer
        if not isinstance(internal_value, int):
            raise TypeError(f"Count options take integer values, not {internal_value}")

        # Value must be reverse-translatable if we have a translation
        if self._translation is not None and not self._translation.can_translate_reverse(internal_value):
            raise ValueError(f"Value {internal_value} is not valid for the translation table")

    def _validate_namespace_value(self, namespace_value: Any):
        # Namespace values are integers
        if not isinstance(namespace_value, int):
            raise TypeError(f"Count options expect integer namespace values, not {namespace_value}")

        # Must be at least 0 (can't have negative count)
        if namespace_value < 0:
            raise ValueError(f"Can't have negative count ({namespace_value})")



from typing import List

from ._Point import Point


class Polygon:
    """
    Represents a polygon in 2D integer Cartesian space.
    """
    def __init__(self, *points: Point):
        self.points: List[Point] = list(points)

    def __iter__(self):
        return iter(self.points)

    def __str__(self):
        return f"[{', '.join(map(str, self.points))}]"

    def is_degenerate(self) -> bool:
        """
        Whether this polygon is degenerate.
        """
        # A polygon with 0, 1 or 2 points is degenerate
        if len(self.points) < 3:
            return True

        # A closed polygon with 3 points really only has 2
        if len(self.points) == 3 and self.is_closed():
            return True

        return False

    def area(self) -> float:
        """
        Calculates the area of the polygon using Greene's theorem.
        Assumes the polygon is not self-intersecting.

        Based on: https://stackoverflow.com/a/451482

        :return:    The area of the polygon.
        """
        return abs(
            sum(
                p1.x * p2.y - p2.x * p1.y
                for p1, p2 in zip(self.points, self.points[1:] + self.points[:1])
            )
        ) / 2

    def simplify(self):
        """
        Simplifies the polygon by removing co-linear/duplicate points.
        """
        # Find all middle co-linear points
        colinear_point_indices: List[int] = []
        for point_index in range(1, len(self.points) - 1):
            if Point.colinear(self.points[point_index - 1],
                              self.points[point_index],
                              self.points[point_index + 1]):
                colinear_point_indices.append(point_index)

        # Remove all co-linear points
        for point_index in reversed(colinear_point_indices):
            self.points.pop(point_index)

        # Find all duplicate points
        duplicate_point_indices: List[int] = []
        for point_index in range(1, len(self.points)):
            if self.points[point_index] == self.points[point_index - 1]:
                duplicate_point_indices.append(point_index)

        # Remove all duplicate points
        for point_index in reversed(duplicate_point_indices):
            self.points.pop(point_index)

    def is_closed(self) -> bool:
        """
        Whether the first and last points are the same.

        :return:    Whether the polygon is closed.
        """
        return len(self.points) > 0 and self.points[0] == self.points[-1]

    def close(self):
        """
        Closes this polygon, adding a copy of the first point
        as the last, if it's not there already.
        """
        # Can't close a polygon with no points
        if len(self.points) == 0:
            return

        # No need to do anything if already closed
        if self.is_closed():
            return

        self.points.append(self.points[0])



from typing import Tuple, Any, Dict

from wai.test import AbstractTest
from wai.test.decorators import Test, ExceptionTest

from wai.common.meta import with_metadata, has_metadata, get_metadata


# Test class for holding meta-data
class TestClass:
    pass


class MetadataTest(AbstractTest):
    @classmethod
    def subject_type(cls):
        return with_metadata

    @classmethod
    def common_arguments(cls) -> Tuple[Tuple[object, str, int], Dict[str, Any]]:
        return (TestClass(), "test_key", 33), {}

    @Test
    def default(self, subject: TestClass):
        self.assertTrue(has_metadata(subject, "test_key"))
        self.assertEqual(get_metadata(subject, "test_key"), 33)

    @ExceptionTest(KeyError)
    def get_missing_key(self, subject: TestClass):
        get_metadata(subject, "missing")
